"""Module for the Self-Adaptive PINN solver."""

from copy import deepcopy
import torch

from ...utils import check_consistency
from ...problem import InverseProblem
from ..solver import MultiSolverInterface
from .pinn_interface import PINNInterface


class Weights(torch.nn.Module):
    """
    Implementation of the mask model for the self-adaptive weights of the
    :class:`SelfAdaptivePINN` solver.
    """

    def __init__(self, func):
        """
        Initialization of the :class:`Weights` class.

        :param torch.nn.Module func: the mask model.
        """
        super().__init__()
        check_consistency(func, torch.nn.Module)
        self.sa_weights = torch.nn.Parameter(torch.Tensor())
        self.func = func

    def forward(self):
        """
        Forward pass implementation for the mask module.

        :return: evaluation of self adaptive weights through the mask.
        :rtype: torch.Tensor
        """
        return self.func(self.sa_weights)


class SelfAdaptivePINN(PINNInterface, MultiSolverInterface):
    r"""
    Self-Adaptive Physics-Informed Neural Network (SelfAdaptivePINN) solver
    class. This class implements the Self-Adaptive Physics-Informed Neural
    Network solver, using a user specified ``model`` to solve a specific
    ``problem``. It can be used to solve both forward and inverse problems.

    The Self-Adapive Physics-Informed Neural Network solver aims to find the
    solution :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m` of a differential
    problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}
    
    integrating pointwise loss evaluation using a mask :math:m and self-adaptive
    weights, which allow the model to focus on regions of the domain where the
    residual is higher.

    The loss function to solve the problem is

    .. math::

        \mathcal{L}_{\rm{problem}} = \frac{1}{N} \sum_{i=1}^{N_\Omega} m
        \left( \lambda_{\Omega}^{i} \right) \mathcal{L} \left( \mathcal{A}
        [\mathbf{u}](\mathbf{x}) \right) + \frac{1}{N} 
        \sum_{i=1}^{N_{\partial\Omega}}
        m \left( \lambda_{\partial\Omega}^{i} \right) \mathcal{L} 
        \left( \mathcal{B}[\mathbf{u}](\mathbf{x})
        \right),
    
    denoting the self adaptive weights as
    :math:`\lambda_{\Omega}^1, \dots, \lambda_{\Omega}^{N_\Omega}` and
    :math:`\lambda_{\partial \Omega}^1, \dots, 
    \lambda_{\Omega}^{N_\partial \Omega}`
    for :math:`\Omega` and :math:`\partial \Omega`, respectively.

    The Self-Adaptive Physics-Informed Neural Network solver identifies the
    solution and appropriate self adaptive weights by solving the following
    optimization problem:

    .. math::

        \min_{w} \max_{\lambda_{\Omega}^k, \lambda_{\partial \Omega}^s}
        \mathcal{L} ,
    
    where :math:`w` denotes the network parameters, and :math:`\mathcal{L}` is a
    specific loss function, , typically the MSE:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.

    .. seealso::
        **Original reference**: McClenny, Levi D., and Ulisses M. Braga-Neto.
        *Self-adaptive physics-informed neural networks.*
        Journal of Computational Physics 474 (2023): 111722.
        DOI: `10.1016/j.jcp.2022.111722
        <https://doi.org/10.1016/j.jcp.2022.111722>`_.
    """

    def __init__(
        self,
        problem,
        model,
        weight_function=torch.nn.Sigmoid(),
        optimizer_model=None,
        optimizer_weights=None,
        scheduler_model=None,
        scheduler_weights=None,
        weighting=None,
        loss=None,
    ):
        """
        Initialization of the :class:`SelfAdaptivePINN` class.

        :param AbstractProblem problem: The problem to be solved.
        :param torch.nn.Module model: The model to be used.
        :param torch.nn.Module weight_function: The Self-Adaptive mask model.
            Default is ``torch.nn.Sigmoid()``.
        :param Optimizer optimizer_model: The optimizer of the ``model``.
            If ``None``, the :class:`torch.optim.Adam` optimizer is used.
            Default is ``None``.
        :param Optimizer optimizer_weights: The optimizer of the
            ``weight_function``.
            If ``None``, the :class:`torch.optim.Adam` optimizer is used.
            Default is ``None``.
        :param Scheduler scheduler_model: Learning rate scheduler for the
            ``model``.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param Scheduler scheduler_weights: Learning rate scheduler for the
            ``weight_function``.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param WeightingInterface weighting: The weighting schema to be used.
            If ``None``, no weighting schema is used. Default is ``None``.
        :param torch.nn.Module loss: The loss function to be minimized.
            If ``None``, the :class:`torch.nn.MSELoss` loss is used.
            Default is `None`.
        """
        # check consistency weitghs_function
        check_consistency(weight_function, torch.nn.Module)

        # create models for weights
        weights_dict = {}
        for condition_name in problem.conditions:
            weights_dict[condition_name] = Weights(weight_function)
        weights_dict = torch.nn.ModuleDict(weights_dict)

        super().__init__(
            models=[model, weights_dict],
            problem=problem,
            optimizers=[optimizer_model, optimizer_weights],
            schedulers=[scheduler_model, scheduler_weights],
            weighting=weighting,
            loss=loss,
        )

        self._vectorial_loss = deepcopy(self.loss)
        self._vectorial_loss.reduction = "none"

    def forward(self, x):
        """
        Forward pass.

        :param LabelTensor x: Input tensor.
        :return: The output of the neural network.
        :rtype: LabelTensor
        """
        return self.model(x)

    def training_step(self, batch):
        """
        Solver training step, overridden to perform manual optimization.

        :param list[tuple[str, dict]] batch: A batch of data. Each element is a
            tuple containing a condition name and a dictionary of points.
        :return: The aggregated loss.
        :rtype: LabelTensor
        """
        # Weights optimization
        self.optimizer_weights.instance.zero_grad()
        loss = super().training_step(batch)
        self.manual_backward(-loss)
        self.optimizer_weights.instance.step()
        self.scheduler_weights.instance.step()

        # Model optimization
        self.optimizer_model.instance.zero_grad()
        loss = super().training_step(batch)
        self.manual_backward(loss)
        self.optimizer_model.instance.step()
        self.scheduler_model.instance.step()

        return loss

    def configure_optimizers(self):
        """
        Optimizer configuration.

        :return: The optimizers and the schedulers
        :rtype: tuple[list[Optimizer], list[Scheduler]]
        """
        # If the problem is an InverseProblem, add the unknown parameters
        # to the parameters to be optimized
        self.optimizer_model.hook(self.model.parameters())
        self.optimizer_weights.hook(self.weights_dict.parameters())
        if isinstance(self.problem, InverseProblem):
            self.optimizer_model.instance.add_param_group(
                {
                    "params": [
                        self._params[var]
                        for var in self.problem.unknown_variables
                    ]
                }
            )
        self.scheduler_model.hook(self.optimizer_model)
        self.scheduler_weights.hook(self.optimizer_weights)
        return (
            [self.optimizer_model.instance, self.optimizer_weights.instance],
            [self.scheduler_model.instance, self.scheduler_weights.instance],
        )

    def on_train_start(self):
        """
        This method is called at the start of the training process to set the
        self-adaptive weights as parameters of the mask model.

        :raises NotImplementedError: If the batch size is not ``None``.
        """
        if self.trainer.batch_size is not None:
            raise NotImplementedError(
                "SelfAdaptivePINN only works with full "
                "batch size, set batch_size=None inside "
                "the Trainer to use the solver."
            )
        device = torch.device(
            self.trainer._accelerator_connector._accelerator_flag
        )

        # Initialize the self adaptive weights only for training points
        for (
            condition_name,
            tensor,
        ) in self.trainer.data_module.train_dataset.input.items():
            self.weights_dict[condition_name].sa_weights.data = torch.rand(
                (tensor.shape[0], 1), device=device
            )
        return super().on_train_start()

    def on_load_checkpoint(self, checkpoint):
        """
        Override of the Pytorch Lightning ``on_load_checkpoint`` method to
        handle checkpoints for Self-Adaptive Weights. This method should not be
        overridden, if not intentionally.

        :param dict checkpoint: Pytorch Lightning checkpoint dict.
        """
        # First initialize self-adaptive weights with correct shape,
        # then load the values from the checkpoint.
        for condition_name, _ in self.problem.input_pts.items():
            shape = checkpoint["state_dict"][
                f"_pina_models.1.{condition_name}.sa_weights"
            ].shape
            self.weights_dict[condition_name].sa_weights.data = torch.rand(
                shape
            )
        return super().on_load_checkpoint(checkpoint)

    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the physics-informed solver based on the
        provided samples and equation.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation.
        :return: The computed physics loss.
        :rtype: LabelTensor
        """
        residual = self.compute_residual(samples, equation)
        weights = self.weights_dict[self.current_condition_name].forward()
        loss_value = self._vectorial_loss(
            torch.zeros_like(residual, requires_grad=True), residual
        )
        return self._vect_to_scalar(weights * loss_value)

    def loss_data(self, input, target):
        """
        Compute the data loss for the PINN solver by evaluating the loss
        between the network's output and the true solution. This method should
        not be overridden, if not intentionally.

        :param input: The input to the neural network.
        :type input: LabelTensor
        :param target: The target to compare with the network's output.
        :type target: LabelTensor
        :return: The supervised loss, averaged over the number of observations.
        :rtype: LabelTensor
        """
        return self._loss_fn(self.forward(input), target)

    def _vect_to_scalar(self, loss_value):
        """
        Computation of the scalar loss.

        :param LabelTensor loss_value: the tensor of pointwise losses.
        :raises RuntimeError: If the loss reduction is not ``mean`` or ``sum``.
        :return: The computed scalar loss.
        :rtype: LabelTensor
        """
        if self.loss.reduction == "mean":
            ret = torch.mean(loss_value)
        elif self.loss.reduction == "sum":
            ret = torch.sum(loss_value)
        else:
            raise RuntimeError(
                f"Invalid reduction, got {self.loss.reduction} "
                "but expected mean or sum."
            )
        return ret

    @property
    def model(self):
        """
        The model.

        :return: The model.
        :rtype: torch.nn.Module
        """
        return self.models[0]

    @property
    def weights_dict(self):
        """
        The self-adaptive weights.

        :return: The self-adaptive weights.
        :rtype: torch.nn.Module
        """
        return self.models[1]

    @property
    def scheduler_model(self):
        """
        The scheduler associated to the model.

        :return: The scheduler for the model.
        :rtype: Scheduler
        """
        return self.schedulers[0]

    @property
    def scheduler_weights(self):
        """
        The scheduler associated to the mask model.

        :return: The scheduler for the mask model.
        :rtype: Scheduler
        """
        return self.schedulers[1]

    @property
    def optimizer_model(self):
        """
        Returns the optimizer associated to the model.

        :return: The optimizer for the model.
        :rtype: Optimizer
        """
        return self.optimizers[0]

    @property
    def optimizer_weights(self):
        """
        The optimizer associated to the mask model.

        :return: The optimizer for the mask model.
        :rtype: Optimizer
        """
        return self.optimizers[1]



import torch
import pytest
from pina.optim import TorchOptimizer, TorchScheduler

opt_list = [
    torch.optim.Adam,
    torch.optim.AdamW,
    torch.optim.SGD,
    torch.optim.RMSprop,
]

sch_list = [torch.optim.lr_scheduler.ConstantLR]


@pytest.mark.parametrize("scheduler_class", sch_list)
def test_constructor(scheduler_class):
    TorchScheduler(scheduler_class)


@pytest.mark.parametrize("optimizer_class", opt_list)
@pytest.mark.parametrize("scheduler_class", sch_list)
def test_hook(optimizer_class, scheduler_class):
    opt = TorchOptimizer(optimizer_class, lr=1e-3)
    opt.hook(torch.nn.Linear(10, 10).parameters())
    sch = TorchScheduler(scheduler_class)
    sch.hook(opt)



"""Module for the TimeDependentProblem class."""

from abc import abstractmethod

from .abstract_problem import AbstractProblem


class TimeDependentProblem(AbstractProblem):
    """
    Class for defining time-dependent problems, where the system's behavior
    changes with respect to time.
    """

    @abstractmethod
    def temporal_domain(self):
        """
        The temporal domain of the problem.
        """

    @property
    def temporal_variable(self):
        """
        Get the time variable of the problem.

        :return: The time variable of the problem.
        :rtype: list[str]
        """
        return self.temporal_domain.variables



"""Module to build Graph objects and perform operations on them."""

import torch
from torch_geometric.data import Data, Batch
from torch_geometric.utils import to_undirected
from torch_geometric.utils.loop import remove_self_loops
from .label_tensor import LabelTensor
from .utils import check_consistency, is_function


class Graph(Data):
    """
    Extends :class:`~torch_geometric.data.Data` class to include additional
    checks and functionlities.
    """

    def __new__(
        cls,
        **kwargs,
    ):
        """
        Create a new instance of the :class:`~pina.graph.Graph` class by
        checking the consistency of the input data and storing the attributes.

        :param dict kwargs: Parameters used to initialize the
            :class:`~pina.graph.Graph` object.
        :return: A new instance of the :class:`~pina.graph.Graph` class.
        :rtype: Graph
        """
        # create class instance
        instance = Data.__new__(cls)

        # check the consistency of types defined in __init__, the others are not
        # checked (as in pyg Data object)
        instance._check_type_consistency(**kwargs)

        return instance

    def __init__(
        self,
        x=None,
        edge_index=None,
        pos=None,
        edge_attr=None,
        undirected=False,
        **kwargs,
    ):
        """
        Initialize the object by setting the node features, edge index,
        edge attributes, and positions. The edge index is preprocessed to make
        the graph undirected if required. For more details, see the
        :meth:`torch_geometric.data.Data`

        :param x: Optional tensor of node features ``(N, F)`` where ``F`` is the
            number of features per node.
        :type x: torch.Tensor, LabelTensor
        :param torch.Tensor edge_index: A tensor of shape ``(2, E)``
            representing the indices of the graph's edges.
        :param pos: A tensor of shape ``(N, D)`` representing the positions of
            ``N`` points in ``D``-dimensional space.
        :type pos: torch.Tensor | LabelTensor
        :param edge_attr: Optional tensor of edge_featured ``(E, F')`` where
            ``F'`` is the number of edge features
        :type edge_attr: torch.Tensor | LabelTensor
        :param bool undirected: Whether to make the graph undirected
        :param dict kwargs: Additional keyword arguments passed to the
            :class:`~torch_geometric.data.Data` class constructor.
        """
        # preprocessing
        self._preprocess_edge_index(edge_index, undirected)

        # calling init
        super().__init__(
            x=x, edge_index=edge_index, edge_attr=edge_attr, pos=pos, **kwargs
        )

    def _check_type_consistency(self, **kwargs):
        """
        Check the consistency of the types of the input data.

        :param dict kwargs: Attributes to be checked for consistency.
        """
        # default types, specified in cls.__new__, by default they are Nont
        # if specified in **kwargs they get override
        x, pos, edge_index, edge_attr = None, None, None, None
        if "pos" in kwargs:
            pos = kwargs["pos"]
            self._check_pos_consistency(pos)
        if "edge_index" in kwargs:
            edge_index = kwargs["edge_index"]
            self._check_edge_index_consistency(edge_index)
        if "x" in kwargs:
            x = kwargs["x"]
            self._check_x_consistency(x, pos)
        if "edge_attr" in kwargs:
            edge_attr = kwargs["edge_attr"]
            self._check_edge_attr_consistency(edge_attr, edge_index)
        if "undirected" in kwargs:
            undirected = kwargs["undirected"]
            check_consistency(undirected, bool)

    @staticmethod
    def _check_pos_consistency(pos):
        """
        Check if the position tensor is consistent.
        :param torch.Tensor pos: The position tensor.
        :raises ValueError: If the position tensor is not consistent.
        """
        if pos is not None:
            check_consistency(pos, (torch.Tensor, LabelTensor))
            if pos.ndim != 2:
                raise ValueError("pos must be a 2D tensor.")

    @staticmethod
    def _check_edge_index_consistency(edge_index):
        """
        Check if the edge index is consistent.

        :param torch.Tensor edge_index: The edge index tensor.
        :raises ValueError: If the edge index tensor is not consistent.
        """
        check_consistency(edge_index, (torch.Tensor, LabelTensor))
        if edge_index.ndim != 2:
            raise ValueError("edge_index must be a 2D tensor.")
        if edge_index.size(0) != 2:
            raise ValueError("edge_index must have shape [2, num_edges].")

    @staticmethod
    def _check_edge_attr_consistency(edge_attr, edge_index):
        """
        Check if the edge attribute tensor is consistent in type and shape
        with the edge index.

        :param edge_attr: The edge attribute tensor.
        :type edge_attr: torch.Tensor | LabelTensor
        :param torch.Tensor edge_index: The edge index tensor.
        :raises ValueError: If the edge attribute tensor is not consistent.
        """
        if edge_attr is not None:
            check_consistency(edge_attr, (torch.Tensor, LabelTensor))
            if edge_attr.ndim != 2:
                raise ValueError("edge_attr must be a 2D tensor.")
            if edge_attr.size(0) != edge_index.size(1):
                raise ValueError(
                    "edge_attr must have shape "
                    "[num_edges, num_edge_features], expected "
                    f"num_edges {edge_index.size(1)} "
                    f"got {edge_attr.size(0)}."
                )

    @staticmethod
    def _check_x_consistency(x, pos=None):
        """
        Check if the input tensor x is consistent with the position tensor
        `pos`.

        :param x: The input tensor.
        :type x: torch.Tensor | LabelTensor
        :param pos: The position tensor.
        :type pos: torch.Tensor | LabelTensor
        :raises ValueError: If the input tensor is not consistent.
        """
        if x is not None:
            check_consistency(x, (torch.Tensor, LabelTensor))
            if x.ndim != 2:
                raise ValueError("x must be a 2D tensor.")
            if pos is not None:
                if x.size(0) != pos.size(0):
                    raise ValueError("Inconsistent number of nodes.")

    @staticmethod
    def _preprocess_edge_index(edge_index, undirected):
        """
        Preprocess the edge index to make the graph undirected (if required).

        :param torch.Tensor edge_index: The edge index.
        :param bool undirected: Whether the graph is undirected.
        :return: The preprocessed edge index.
        :rtype: torch.Tensor
        """
        if undirected:
            edge_index = to_undirected(edge_index)
        return edge_index

    def extract(self, labels, attr="x"):
        """
        Perform extraction of labels from the attribute specified by `attr`.

        :param labels: Labels to extract
        :type labels: list[str] | tuple[str] | str | dict
        :return: Batch object with extraction performed on x
        :rtype: PinaBatch
        """
        # Extract labels from LabelTensor object
        tensor = getattr(self, attr).extract(labels)
        # Set the extracted tensor as the new attribute
        setattr(self, attr, tensor)
        return self


class GraphBuilder:
    """
    A class that allows an easy definition of :class:`Graph` instances.
    """

    def __new__(
        cls,
        pos,
        edge_index,
        x=None,
        edge_attr=False,
        custom_edge_func=None,
        loop=True,
        **kwargs,
    ):
        """
        Compute the edge attributes and create a new instance of the
        :class:`~pina.graph.Graph` class.

        :param pos: A tensor of shape ``(N, D)`` representing the positions of
            ``N`` points in ``D``-dimensional space.
        :type pos: torch.Tensor or LabelTensor
        :param edge_index: A tensor of shape ``(2, E)`` representing the indices
            of the graph's edges.
        :type edge_index: torch.Tensor
        :param x: Optional tensor of node features of shape ``(N, F)``, where
            ``F`` is the number of features per node.
        :type x: torch.Tensor | LabelTensor, optional
        :param bool edge_attr: Whether to compute the edge attributes.
        :param custom_edge_func: A custom function to compute edge attributes.
            If provided, overrides ``edge_attr``.
        :type custom_edge_func: Callable, optional
        :param bool loop: Whether to include self-loops.
        :param kwargs: Additional keyword arguments passed to the
            :class:`~pina.graph.Graph` class constructor.
        :return: A :class:`~pina.graph.Graph` instance constructed using the
            provided information.
        :rtype: Graph
        """
        if not loop:
            edge_index = remove_self_loops(edge_index)[0]
        edge_attr = cls._create_edge_attr(
            pos, edge_index, edge_attr, custom_edge_func or cls._build_edge_attr
        )
        return Graph(
            x=x,
            edge_index=edge_index,
            edge_attr=edge_attr,
            pos=pos,
            **kwargs,
        )

    @staticmethod
    def _create_edge_attr(pos, edge_index, edge_attr, func):
        """
        Create the edge attributes based on the input parameters.

        :param pos: Positions of the points.
        :type pos: torch.Tensor | LabelTensor
        :param torch.Tensor edge_index: Edge indices.
        :param bool edge_attr: Whether to compute the edge attributes.
        :param Callable func: Function to compute the edge attributes.
        :raises ValueError: If ``func`` is not a function.
        :return: The edge attributes.
        :rtype: torch.Tensor | LabelTensor | None
        """
        check_consistency(edge_attr, bool)
        if edge_attr:
            if is_function(func):
                return func(pos, edge_index)
            raise ValueError("custom_edge_func must be a function.")
        return None

    @staticmethod
    def _build_edge_attr(pos, edge_index):
        """
        Default function to compute the edge attributes.

        :param pos: Positions of the points.
        :type pos: torch.Tensor | LabelTensor
        :param torch.Tensor edge_index: Edge indices.
        :return: The edge attributes.
        :rtype: torch.Tensor
        """
        return (
            (pos[edge_index[0]] - pos[edge_index[1]])
            .abs()
            .as_subclass(torch.Tensor)
        )


class RadiusGraph(GraphBuilder):
    """
    Extends the :class:`~pina.graph.GraphBuilder` class to compute
    ``edge_index`` based on a radius. Each point is connected to all the points
    within the radius.
    """

    def __new__(cls, pos, radius, **kwargs):
        """
        Instantiate the :class:`~pina.graph.Graph` class by computing the
        ``edge_index`` based on the radius provided.

        :param pos: A tensor of shape ``(N, D)`` representing the positions of
            ``N`` points in ``D``-dimensional space.
        :type pos: torch.Tensor | LabelTensor
        :param float radius: The radius within which points are connected.
        :param dict kwargs: The additional keyword arguments to be passed to
            :class:`GraphBuilder` and :class:`Graph` classes.
        :return: A :class:`~pina.graph.Graph` instance with the computed
            ``edge_index``.
        :rtype: Graph
        """
        edge_index = cls.compute_radius_graph(pos, radius)
        return super().__new__(cls, pos=pos, edge_index=edge_index, **kwargs)

    @staticmethod
    def compute_radius_graph(points, radius):
        """
        Computes the ``edge_index`` based on the radius. Each point is connected
        to all the points within the radius.

        :param points: A tensor of shape ``(N, D)`` representing the positions
            of ``N`` points in ``D``-dimensional space.
        :type points: torch.Tensor | LabelTensor
        :param float radius: The radius within which points are connected.
        :return: A tensor of shape ``(2, E)``, with ``E`` number of edges,
            representing the edge indices of the graph.
        :rtype: torch.Tensor
        """
        dist = torch.cdist(points, points, p=2)
        return (
            torch.nonzero(dist <= radius, as_tuple=False)
            .t()
            .as_subclass(torch.Tensor)
        )


class KNNGraph(GraphBuilder):
    """
    Extends the :class:`~pina.graph.GraphBuilder` class to compute
    ``edge_index`` based on a K-nearest neighbors algorithm.
    """

    def __new__(cls, pos, neighbours, **kwargs):
        """
        Instantiate the :class:`~pina.graph.Graph` class by computing the
        ``edge_index`` based on the K-nearest neighbors algorithm.

        :param pos: A tensor of shape ``(N, D)`` representing the positions of
            ``N`` points in ``D``-dimensional space.
        :type pos: torch.Tensor | LabelTensor
        :param int neighbours: The number of nearest neighbors to consider when
            building the graph.
        :param dict kwargs: The additional keyword arguments to be passed to
            :class:`GraphBuilder` and :class:`Graph` classes.

        :return: A :class:`~pina.graph.Graph` instance with the computed
            ``edge_index``.
        :rtype: Graph
        """

        edge_index = cls.compute_knn_graph(pos, neighbours)
        return super().__new__(cls, pos=pos, edge_index=edge_index, **kwargs)

    @staticmethod
    def compute_knn_graph(points, neighbours):
        """
        Computes the ``edge_index`` based on the K-nearest neighbors algorithm.

        :param points: A tensor of shape ``(N, D)`` representing the positions
            of ``N`` points in ``D``-dimensional space.
        :type points: torch.Tensor | LabelTensor
        :param int neighbours: The number of nearest neighbors to consider when
            building the graph.
        :return: A tensor of shape ``(2, E)``, with ``E`` number of edges,
            representing the edge indices of the graph.
        :rtype: torch.Tensor
        """
        dist = torch.cdist(points, points, p=2)
        knn_indices = torch.topk(dist, k=neighbours, largest=False).indices
        row = torch.arange(points.size(0)).repeat_interleave(neighbours)
        col = knn_indices.flatten()
        return torch.stack([row, col], dim=0).as_subclass(torch.Tensor)


class LabelBatch(Batch):
    """
    Extends the :class:`~torch_geometric.data.Batch` class to include
    :class:`~pina.label_tensor.LabelTensor` objects.
    """

    @classmethod
    def from_data_list(cls, data_list):
        """
        Create a Batch object from a list of :class:`~torch_geometric.data.Data`
        or :class:`~pina.graph.Graph` objects.

        :param data_list: List of :class:`~torch_geometric.data.Data` or
            :class:`~pina.graph.Graph` objects.
        :type data_list: list[Data] | list[Graph]
        :return: A :class:`~torch_geometric.data.Batch` object containing
            the input data.
        :rtype: :class:`~torch_geometric.data.Batch`
        """
        # Store the labels of Data/Graph objects (all data have the same labels)
        # If the data do not contain labels, labels is an empty dictionary,
        # therefore the labels are not stored
        labels = {
            k: v.labels
            for k, v in data_list[0].items()
            if isinstance(v, LabelTensor)
        }

        # Create a Batch object from the list of Data objects
        batch = super().from_data_list(data_list)

        # Put the labels back in the Batch object
        for k, v in labels.items():
            batch[k].labels = v
        return batch



"""Module for spectral convolution blocks."""

import torch
from torch import nn
from ...utils import check_consistency


######## 1D Spectral Convolution ###########
class SpectralConvBlock1D(nn.Module):
    """
    Spectral Convolution Block for one-dimensional tensors.

    This class computes the spectral convolution of the input with a linear
    kernel in the fourier space, and then it maps the input back to the physical
    space.
    The block expects an input of size [``batch``, ``input_numb_fields``, ``N``]
    and returns an output of size [``batch``, ``output_numb_fields``, ``N``].
    """

    def __init__(self, input_numb_fields, output_numb_fields, n_modes):
        r"""
        Initialization of the :class:`SpectralConvBlock1D` class.

        :param int input_numb_fields: The number of channels for the input.
        :param int output_numb_fields: The number of channels for the output.
        :param int n_modes: The number of modes to select for each dimension.
            It must be at most equal to :math:`\floor(Nx/2)+1`.
        """
        super().__init__()

        # check type consistency
        check_consistency(input_numb_fields, int)
        check_consistency(output_numb_fields, int)

        # assign variables
        self._modes = n_modes
        self._input_channels = input_numb_fields
        self._output_channels = output_numb_fields

        # scaling factor
        scale = 1.0 / (self._input_channels * self._output_channels)
        self._weights = nn.Parameter(
            scale
            * torch.rand(
                self._input_channels,
                self._output_channels,
                self._modes,
                dtype=torch.cfloat,
            )
        )

    def _compute_mult1d(self, input, weights):
        """
        Compute the matrix multiplication of the input and the linear kernel
        weights.

        :param torch.Tensor input: The input tensor. Expected of size
            [``batch``, ``input_numb_fields``, ``N``].
        :param torch.Tensor weights: The kernel weights. Expected of size
            [``input_numb_fields``, ``output_numb_fields``, ``N``].
        :return: The result of the matrix multiplication.
        :rtype: torch.Tensor
        """
        return torch.einsum("bix,iox->box", input, weights)

    def forward(self, x):
        """
        Forward pass.

        :param torch.Tensor x: The input tensor. Expected of size
            [``batch``, ``input_numb_fields``, ``N``].
        :return: The input tensor. Expected of size
            [``batch``, ``output_numb_fields``, ``N``].
        :rtype: torch.Tensor
        """
        batch_size = x.shape[0]

        # Compute Fourier transform of the input
        x_ft = torch.fft.rfft(x)

        # Multiply relevant Fourier modes
        out_ft = torch.zeros(
            batch_size,
            self._output_channels,
            x.size(-1) // 2 + 1,
            device=x.device,
            dtype=torch.cfloat,
        )
        out_ft[:, :, : self._modes] = self._compute_mult1d(
            x_ft[:, :, : self._modes], self._weights
        )

        # Return to physical space
        return torch.fft.irfft(out_ft, n=x.size(-1))


######## 2D Spectral Convolution ###########
class SpectralConvBlock2D(nn.Module):
    """
    Spectral Convolution Block for two-dimensional tensors.

    This class computes the spectral convolution of the input with a linear
    kernel in the fourier space, and then it maps the input back to the physical
    space.
    The block expects an input of size
    [``batch``, ``input_numb_fields``, ``Nx``, ``Ny``]
    and returns an output of size
    [``batch``, ``output_numb_fields``, ``Nx``, ``Ny``].
    """

    def __init__(self, input_numb_fields, output_numb_fields, n_modes):
        r"""
        Initialization of the :class:`SpectralConvBlock2D` class.

        :param int input_numb_fields: The number of channels for the input.
        :param int output_numb_fields: The number of channels for the output.
        :param n_modes: The number of modes to select for each dimension.
            It must be at most equal to :math:`\floor(Nx/2)+1`,
            :math:`\floor(Ny/2)+1`.
        :type n_modes: list[int] | tuple[int]
        :raises ValueError: If the number of modes is not consistent.
        :raises ValueError: If the number of modes is not a list or tuple.
        """
        super().__init__()

        # check type consistency
        check_consistency(input_numb_fields, int)
        check_consistency(output_numb_fields, int)
        check_consistency(n_modes, int)
        if isinstance(n_modes, (tuple, list)):
            if len(n_modes) != 2:
                raise ValueError(
                    "Expected n_modes to be a list or tuple of len two, "
                    "with each entry corresponding to the number of modes "
                    "for each dimension "
                )
        elif isinstance(n_modes, int):
            n_modes = [n_modes] * 2
        else:
            raise ValueError(
                "Expected n_modes to be a list or tuple of len two, "
                "with each entry corresponding to the number of modes "
                "for each dimension; or an int value representing the "
                "number of modes for all dimensions"
            )

        # assign variables
        self._modes = n_modes
        self._input_channels = input_numb_fields
        self._output_channels = output_numb_fields

        # scaling factor
        scale = 1.0 / (self._input_channels * self._output_channels)
        self._weights1 = nn.Parameter(
            scale
            * torch.rand(
                self._input_channels,
                self._output_channels,
                self._modes[0],
                self._modes[1],
                dtype=torch.cfloat,
            )
        )
        self._weights2 = nn.Parameter(
            scale
            * torch.rand(
                self._input_channels,
                self._output_channels,
                self._modes[0],
                self._modes[1],
                dtype=torch.cfloat,
            )
        )

    def _compute_mult2d(self, input, weights):
        """
        Compute the matrix multiplication of the input and the linear kernel
        weights.

        :param torch.Tensor input: The input tensor. Expected of size
            [``batch``, ``input_numb_fields``, ``Nx``, ``Ny``].
        :param torch.Tensor weights: The kernel weights. Expected of size
            [``input_numb_fields``, ``output_numb_fields``, ``Nx``, ``Ny``].
        :return: The result of the matrix multiplication.
        :rtype: torch.Tensor
        """
        return torch.einsum("bixy,ioxy->boxy", input, weights)

    def forward(self, x):
        """
        Forward pass.

        :param torch.Tensor x: The input tensor. Expected of size
            [``batch``, ``input_numb_fields``, ``Nx``, ``Ny``].
        :return: The input tensor. Expected of size
            [``batch``, ``output_numb_fields``, ``Nx``, ``Ny``].
        :rtype: torch.Tensor
        """

        batch_size = x.shape[0]

        # Compute Fourier transform of the input
        x_ft = torch.fft.rfft2(x)

        # Multiply relevant Fourier modes
        out_ft = torch.zeros(
            batch_size,
            self._output_channels,
            x.size(-2),
            x.size(-1) // 2 + 1,
            device=x.device,
            dtype=torch.cfloat,
        )
        out_ft[:, :, : self._modes[0], : self._modes[1]] = self._compute_mult2d(
            x_ft[:, :, : self._modes[0], : self._modes[1]], self._weights1
        )
        out_ft[:, :, -self._modes[0] :, : self._modes[1] :] = (
            self._compute_mult2d(
                x_ft[:, :, -self._modes[0] :, : self._modes[1]], self._weights2
            )
        )

        # Return to physical space
        return torch.fft.irfft2(out_ft, s=(x.size(-2), x.size(-1)))


######## 3D Spectral Convolution ###########
class SpectralConvBlock3D(nn.Module):
    """
    Spectral Convolution Block for three-dimensional tensors.

    This class computes the spectral convolution of the input with a linear
    kernel in the fourier space, and then it maps the input back to the physical
    space.
    The block expects an input of size
    [``batch``, ``input_numb_fields``, ``Nx``, ``Ny``, ``Nz``]
    and returns an output of size
    [``batch``, ``output_numb_fields``, ``Nx``, ``Ny``, ``Nz``].
    """

    def __init__(self, input_numb_fields, output_numb_fields, n_modes):
        r"""
        Initialization of the :class:`SpectralConvBlock3D` class.

        :param int input_numb_fields: The number of channels for the input.
        :param int output_numb_fields: The number of channels for the output.
        :param n_modes: The number of modes to select for each dimension.
            It must be at most equal to :math:`\floor(Nx/2)+1`,
            :math:`\floor(Ny/2)+1`, :math:`\floor(Nz/2)+1`.
        :type n_modes: list[int] | tuple[int]
        :raises ValueError: If the number of modes is not consistent.
        :raises ValueError: If the number of modes is not a list or tuple.
        """
        super().__init__()

        # check type consistency
        check_consistency(input_numb_fields, int)
        check_consistency(output_numb_fields, int)
        check_consistency(n_modes, int)
        if isinstance(n_modes, (tuple, list)):
            if len(n_modes) != 3:
                raise ValueError(
                    "Expected n_modes to be a list or tuple of len three, "
                    "with each entry corresponding to the number of modes "
                    "for each dimension "
                )
        elif isinstance(n_modes, int):
            n_modes = [n_modes] * 3
        else:
            raise ValueError(
                "Expected n_modes to be a list or tuple of len three, "
                "with each entry corresponding to the number of modes "
                "for each dimension; or an int value representing the "
                "number of modes for all dimensions"
            )

        # assign variables
        self._modes = n_modes
        self._input_channels = input_numb_fields
        self._output_channels = output_numb_fields

        # scaling factor
        scale = 1.0 / (self._input_channels * self._output_channels)
        self._weights1 = nn.Parameter(
            scale
            * torch.rand(
                self._input_channels,
                self._output_channels,
                self._modes[0],
                self._modes[1],
                self._modes[2],
                dtype=torch.cfloat,
            )
        )
        self._weights2 = nn.Parameter(
            scale
            * torch.rand(
                self._input_channels,
                self._output_channels,
                self._modes[0],
                self._modes[1],
                self._modes[2],
                dtype=torch.cfloat,
            )
        )
        self._weights3 = nn.Parameter(
            scale
            * torch.rand(
                self._input_channels,
                self._output_channels,
                self._modes[0],
                self._modes[1],
                self._modes[2],
                dtype=torch.cfloat,
            )
        )
        self._weights4 = nn.Parameter(
            scale
            * torch.rand(
                self._input_channels,
                self._output_channels,
                self._modes[0],
                self._modes[1],
                self._modes[2],
                dtype=torch.cfloat,
            )
        )

    def _compute_mult3d(self, input, weights):
        """
        Compute the matrix multiplication of the input and the linear kernel
        weights.

        :param torch.Tensor input: The input tensor. Expected of size
            [``batch``, ``input_numb_fields``, ``Nx``, ``Ny``, ``Nz``].
        :param torch.Tensor weights: The kernel weights. Expected of size
            [``input_numb_fields``, ``output_numb_fields``, ``Nx``, ``Ny``,
            ``Nz``].
        :return: The result of the matrix multiplication.
        :rtype: torch.Tensor
        """
        return torch.einsum("bixyz,ioxyz->boxyz", input, weights)

    def forward(self, x):
        """
        Forward pass.

        :param torch.Tensor x: The input tensor. Expected of size
            [``batch``, ``input_numb_fields``, ``Nx``, ``Ny``, ``Nz``].
        :return: The input tensor. Expected of size
            [``batch``, ``output_numb_fields``, ``Nx``, ``Ny``, ``Nz``].
        :rtype: torch.Tensor
        """

        batch_size = x.shape[0]

        # Compute Fourier transform of the input
        x_ft = torch.fft.rfftn(x, dim=[-3, -2, -1])

        # Multiply relevant Fourier modes
        out_ft = torch.zeros(
            batch_size,
            self._output_channels,
            x.size(-3),
            x.size(-2),
            x.size(-1) // 2 + 1,
            device=x.device,
            dtype=torch.cfloat,
        )

        slice0 = (
            slice(None),
            slice(None),
            slice(self._modes[0]),
            slice(self._modes[1]),
            slice(self._modes[2]),
        )
        out_ft[slice0] = self._compute_mult3d(x_ft[slice0], self._weights1)

        slice1 = (
            slice(None),
            slice(None),
            slice(self._modes[0]),
            slice(-self._modes[1], None),
            slice(self._modes[2]),
        )
        out_ft[slice1] = self._compute_mult3d(x_ft[slice1], self._weights2)

        slice2 = (
            slice(None),
            slice(None),
            slice(-self._modes[0], None),
            slice(self._modes[1]),
            slice(self._modes[2]),
        )
        out_ft[slice2] = self._compute_mult3d(x_ft[slice2], self._weights3)

        slice3 = (
            slice(None),
            slice(None),
            slice(-self._modes[0], None),
            slice(-self._modes[1], None),
            slice(self._modes[2]),
        )
        out_ft[slice3] = self._compute_mult3d(x_ft[slice3], self._weights4)

        # Return to physical space
        return torch.fft.irfftn(out_ft, s=(x.size(-3), x.size(-2), x.size(-1)))



"""Module for the Graph Neural Operator Block class."""

import torch
from torch_geometric.nn import MessagePassing


class GNOBlock(MessagePassing):
    """
    The inner block of the Graph Neural Operator, based on Message Passing.
    """

    def __init__(
        self,
        width,
        edges_features,
        n_layers=2,
        layers=None,
        inner_size=None,
        internal_func=None,
        external_func=None,
    ):
        """
        Initialization of the :class:`GNOBlock` class.

        :param int width: The width of the kernel.
        :param int edge_features: The number of edge features.
        :param int n_layers: The number of kernel layers. Default is ``2``.
        :param layers: A list specifying the number of neurons for each layer
            of the neural network. If not ``None``, it overrides the
            ``inner_size`` and ``n_layers``parameters. Default is ``None``.
        :type layers: list[int] | tuple[int]
        :param int inner_size: The size of the inner layer. Default is ``None``.
        :param torch.nn.Module internal_func: The activation function applied to
            the output of each layer. If ``None``, it uses the
            :class:`torch.nn.Tanh` activation. Default is ``None``.
        :param torch.nn.Module external_func: The activation function applied to
            the output of the block. If ``None``, it uses the
            :class:`torch.nn.Tanh`. activation. Default is ``None``.
        """

        from ...model.feed_forward import FeedForward

        super().__init__(aggr="mean")  # Uses PyG's default aggregation
        self.width = width

        if layers is None and inner_size is None:
            inner_size = width

        self.dense = FeedForward(
            input_dimensions=edges_features,
            output_dimensions=width**2,
            n_layers=n_layers,
            layers=layers,
            inner_size=inner_size,
            func=internal_func,
        )

        self.W = torch.nn.Linear(width, width)
        self.func = external_func()

    def message_and_aggregate(self, edge_index, x, edge_attr):
        """
        Combine messages and perform aggregation.

        :param torch.Tensor edge_index: The edge index.
        :param torch.Tensor x: The node feature matrix.
        :param torch.Tensor edge_attr: The edge features.
        :return: The aggregated messages.
        :rtype: torch.Tensor
        """
        # Edge features are transformed into a matrix of shape
        # [num_edges, width, width]
        x_ = self.dense(edge_attr).view(-1, self.width, self.width)
        # Messages are computed as the product of the edge features
        messages = torch.einsum("bij,bj->bi", x_, x[edge_index[0]])
        # Aggregation is performed using the mean (set in the constructor)
        return self.aggregate(messages, edge_index[1])

    def edge_update(self, edge_attr):
        """
        Update edge features.

        :param torch.Tensor edge_attr: The edge features.
        :return: The updated edge features.
        :rtype: torch.Tensor
        """
        return edge_attr

    def update(self, aggr_out, x):
        """
        Update node features.

        :param torch.Tensor aggr_out: The aggregated messages.
        :param torch.Tensor x: The node feature matrix.
        :return: The updated node features.
        :rtype: torch.Tensor
        """
        return aggr_out + self.W(x)

    def forward(self, x, edge_index, edge_attr):
        """
        Forward pass of the block.

        :param torch.Tensor x: The node features.
        :param torch.Tensor edge_index: The edge indeces.
        :param torch.Tensor edge_attr: The edge features.
        :return: The updated node features.
        :rtype: torch.Tensor
        """
        return self.func(self.propagate(edge_index, x=x, edge_attr=edge_attr))



"""Formulation of the Allen Cahn problem."""

import torch
from ... import Condition
from ...equation import Equation
from ...domain import CartesianDomain
from ...operator import grad, laplacian
from ...problem import SpatialProblem, TimeDependentProblem


def allen_cahn_equation(input_, output_):
    """
    Implementation of the Allen Cahn equation.

    :param LabelTensor input_: Input data of the problem.
    :param LabelTensor output_: Output data of the problem.
    :return: The residual of the Allen Cahn equation.
    :rtype: LabelTensor
    """
    u_t = grad(output_, input_, components=["u"], d=["t"])
    u_xx = laplacian(output_, input_, components=["u"], d=["x"])
    return u_t - 0.0001 * u_xx + 5 * output_**3 - 5 * output_


def initial_condition(input_, output_):
    """
    Definition of the initial condition of the Allen Cahn problem.

    :param LabelTensor input_: Input data of the problem.
    :param LabelTensor output_: Output data of the problem.
    :return: The residual of the initial condition.
    :rtype: LabelTensor
    """
    x = input_.extract("x")
    u_0 = x**2 * torch.cos(torch.pi * x)
    return output_ - u_0


class AllenCahnProblem(TimeDependentProblem, SpatialProblem):
    r"""
    Implementation of the Allen Cahn problem in the spatial interval
    :math:`[-1, 1]` and temporal interval :math:`[0, 1]`.

    .. seealso::
        **Original reference**: Sokratis J. Anagnostopoulos, Juan D. Toscano,
        Nikolaos Stergiopulos, and George E. Karniadakis.
        *Residual-based attention and connection to information
        bottleneck theory in PINNs*.
        Computer Methods in Applied Mechanics and Engineering 421 (2024): 116805
        DOI: `10.1016/
        j.cma.2024.116805 <https://doi.org/10.1016/j.cma.2024.116805>`_.

    :Example:
        >>> problem = AllenCahnProblem()
    """

    output_variables = ["u"]
    spatial_domain = CartesianDomain({"x": [-1, 1]})
    temporal_domain = CartesianDomain({"t": [0, 1]})

    domains = {
        "D": CartesianDomain({"x": [-1, 1], "t": [0, 1]}),
        "t0": CartesianDomain({"x": [-1, 1], "t": 0.0}),
    }

    conditions = {
        "D": Condition(domain="D", equation=Equation(allen_cahn_equation)),
        "t0": Condition(domain="t0", equation=Equation(initial_condition)),
    }



"""Module for the Adaptive Functions."""

import torch
from ..utils import check_consistency
from .adaptive_function_interface import AdaptiveActivationFunctionInterface


class AdaptiveReLU(AdaptiveActivationFunctionInterface):
    r"""
    Adaptive trainable :class:`~torch.nn.ReLU` activation function.

    Given the function :math:`\text{ReLU}:\mathbb{R}^n\rightarrow\mathbb{R}^n`,
    the adaptive function
    :math:`\text{ReLU}_{\text{adaptive}}:\mathbb{R}^n\rightarrow\mathbb{R}^n`
    is defined as:

    .. math::
        \text{ReLU}_{\text{adaptive}}({x})=\alpha\,\text{ReLU}(\beta{x}+\gamma),

    where :math:`\alpha,\,\beta,\,\gamma` are trainable parameters, and the
    ReLU function is defined as:

    .. math::
        \text{ReLU}(x)  = \max(0, x)

    .. seealso::

        **Original reference**: Godfrey, Luke B., and Michael S. Gashler.
        *A continuum among logarithmic, linear, and exponential functions,
        and its potential to improve generalization in neural networks.*
        2015 7th international joint conference on knowledge discovery,
        knowledge engineering and knowledge management (IC3K).
        Vol. 1. IEEE, 2015. DOI: `arXiv preprint arXiv:1602.01321.
        <https://arxiv.org/abs/1602.01321>`_.

        Jagtap, Ameya D., Kenji Kawaguchi, and George Em Karniadakis. *Adaptive
        activation functions accelerate convergence in deep and
        physics-informed neural networks*. Journal of
        Computational Physics 404 (2020): 109136.
        DOI: `JCP 10.1016
        <https://doi.org/10.1016/j.jcp.2019.109136>`_.
    """

    def __init__(self, alpha=None, beta=None, gamma=None, fixed=None):
        super().__init__(alpha, beta, gamma, fixed)
        self._func = torch.nn.ReLU()


class AdaptiveSigmoid(AdaptiveActivationFunctionInterface):
    r"""
    Adaptive trainable :class:`~torch.nn.Sigmoid` activation function.

    Given the function
    :math:`\text{Sigmoid}:\mathbb{R}^n\rightarrow\mathbb{R}^n`,
    the adaptive function
    :math:`\text{Sigmoid}_{\text{adaptive}}:\mathbb{R}^n\rightarrow\mathbb{R}^n`
    is defined as:

    .. math::
        \text{Sigmoid}_{\text{adaptive}}({x})=
        \alpha\,\text{Sigmoid}(\beta{x}+\gamma),

    where :math:`\alpha,\,\beta,\,\gamma` are trainable parameters, and the
    Sigmoid function is defined as:

    .. math::
        \text{Sigmoid}(x) = \frac{1}{1 + \exp(-x)}

    .. seealso::

        **Original reference**: Godfrey, Luke B., and Michael S. Gashler.
        *A continuum among logarithmic, linear, and exponential functions,
        and its potential to improve generalization in neural networks.*
        2015 7th international joint conference on knowledge discovery,
        knowledge engineering and knowledge management (IC3K).
        Vol. 1. IEEE, 2015. DOI: `arXiv preprint arXiv:1602.01321.
        <https://arxiv.org/abs/1602.01321>`_.

        Jagtap, Ameya D., Kenji Kawaguchi, and George Em Karniadakis. *Adaptive
        activation functions accelerate convergence in deep and
        physics-informed neural networks*. Journal of
        Computational Physics 404 (2020): 109136.
        DOI: `JCP 10.1016
        <https://doi.org/10.1016/j.jcp.2019.109136>`_.
    """

    def __init__(self, alpha=None, beta=None, gamma=None, fixed=None):
        super().__init__(alpha, beta, gamma, fixed)
        self._func = torch.nn.Sigmoid()


class AdaptiveTanh(AdaptiveActivationFunctionInterface):
    r"""
    Adaptive trainable :class:`~torch.nn.Tanh` activation function.

    Given the function :math:`\text{Tanh}:\mathbb{R}^n\rightarrow\mathbb{R}^n`,
    the adaptive function
    :math:`\text{Tanh}_{\text{adaptive}}:\mathbb{R}^n\rightarrow\mathbb{R}^n`
    is defined as:

    .. math::
        \text{Tanh}_{\text{adaptive}}({x})=\alpha\,\text{Tanh}(\beta{x}+\gamma),

    where :math:`\alpha,\,\beta,\,\gamma` are trainable parameters, and the
    Tanh function is defined as:

    .. math::
        \text{Tanh}(x) = \frac{\exp(x) - \exp(-x)} {\exp(x) + \exp(-x)}

    .. seealso::

        **Original reference**: Godfrey, Luke B., and Michael S. Gashler.
        *A continuum among logarithmic, linear, and exponential functions,
        and its potential to improve generalization in neural networks.*
        2015 7th international joint conference on knowledge discovery,
        knowledge engineering and knowledge management (IC3K).
        Vol. 1. IEEE, 2015. DOI: `arXiv preprint arXiv:1602.01321.
        <https://arxiv.org/abs/1602.01321>`_.

        Jagtap, Ameya D., Kenji Kawaguchi, and George Em Karniadakis. *Adaptive
        activation functions accelerate convergence in deep and
        physics-informed neural networks*. Journal of
        Computational Physics 404 (2020): 109136.
        DOI: `JCP 10.1016
        <https://doi.org/10.1016/j.jcp.2019.109136>`_.
    """

    def __init__(self, alpha=None, beta=None, gamma=None, fixed=None):
        super().__init__(alpha, beta, gamma, fixed)
        self._func = torch.nn.Tanh()


class AdaptiveSiLU(AdaptiveActivationFunctionInterface):
    r"""
    Adaptive trainable :class:`~torch.nn.SiLU` activation function.

    Given the function :math:`\text{SiLU}:\mathbb{R}^n\rightarrow\mathbb{R}^n`,
    the adaptive function
    :math:`\text{SiLU}_{\text{adaptive}}:\mathbb{R}^n\rightarrow\mathbb{R}^n`
    is defined as:

    .. math::
        \text{SiLU}_{\text{adaptive}}({x})=\alpha\,\text{SiLU}(\beta{x}+\gamma),

    where :math:`\alpha,\,\beta,\,\gamma` are trainable parameters, and the
    SiLU function is defined as:

    .. math::
        \text{SiLU}(x) = x * \sigma(x), \text{where }\sigma(x)
        \text{ is the logistic sigmoid.}

    .. seealso::

        **Original reference**: Godfrey, Luke B., and Michael S. Gashler.
        *A continuum among logarithmic, linear, and exponential functions,
        and its potential to improve generalization in neural networks.*
        2015 7th international joint conference on knowledge discovery,
        knowledge engineering and knowledge management (IC3K).
        Vol. 1. IEEE, 2015. DOI: `arXiv preprint arXiv:1602.01321.
        <https://arxiv.org/abs/1602.01321>`_.

        Jagtap, Ameya D., Kenji Kawaguchi, and George Em Karniadakis. *Adaptive
        activation functions accelerate convergence in deep and
        physics-informed neural networks*. Journal of
        Computational Physics 404 (2020): 109136.
        DOI: `JCP 10.1016
        <https://doi.org/10.1016/j.jcp.2019.109136>`_.
    """

    def __init__(self, alpha=None, beta=None, gamma=None, fixed=None):
        super().__init__(alpha, beta, gamma, fixed)
        self._func = torch.nn.SiLU()


class AdaptiveMish(AdaptiveActivationFunctionInterface):
    r"""
    Adaptive trainable :class:`~torch.nn.Mish` activation function.

    Given the function :math:`\text{Mish}:\mathbb{R}^n\rightarrow\mathbb{R}^n`,
    the adaptive function
    :math:`\text{Mish}_{\text{adaptive}}:\mathbb{R}^n\rightarrow\mathbb{R}^n`
    is defined as:

    .. math::
        \text{Mish}_{\text{adaptive}}({x})=\alpha\,\text{Mish}(\beta{x}+\gamma),

    where :math:`\alpha,\,\beta,\,\gamma` are trainable parameters, and the
    Mish function is defined as:

    .. math::
        \text{Mish}(x) = x * \text{Tanh}(x)

    .. seealso::

        **Original reference**: Godfrey, Luke B., and Michael S. Gashler.
        *A continuum among logarithmic, linear, and exponential functions,
        and its potential to improve generalization in neural networks.*
        2015 7th international joint conference on knowledge discovery,
        knowledge engineering and knowledge management (IC3K).
        Vol. 1. IEEE, 2015. DOI: `arXiv preprint arXiv:1602.01321.
        <https://arxiv.org/abs/1602.01321>`_.

        Jagtap, Ameya D., Kenji Kawaguchi, and George Em Karniadakis. *Adaptive
        activation functions accelerate convergence in deep and
        physics-informed neural networks*. Journal of
        Computational Physics 404 (2020): 109136.
        DOI: `JCP 10.1016
        <https://doi.org/10.1016/j.jcp.2019.109136>`_.
    """

    def __init__(self, alpha=None, beta=None, gamma=None, fixed=None):
        super().__init__(alpha, beta, gamma, fixed)
        self._func = torch.nn.Mish()


class AdaptiveELU(AdaptiveActivationFunctionInterface):
    r"""
    Adaptive trainable :class:`~torch.nn.ELU` activation function.

    Given the function :math:`\text{ELU}:\mathbb{R}^n\rightarrow\mathbb{R}^n`,
    the adaptive function
    :math:`\text{ELU}_{\text{adaptive}}:\mathbb{R}^n\rightarrow\mathbb{R}^n`
    is defined as:

    .. math::
        \text{ELU}_{\text{adaptive}}({x}) = \alpha\,\text{ELU}(\beta{x}+\gamma),

    where :math:`\alpha,\,\beta,\,\gamma` are trainable parameters, and the
    ELU function is defined as:
    
    .. math::
        \text{ELU}(x) = \begin{cases}
        x, & \text{ if  }x > 0\\
        \exp(x) - 1, & \text{ if  }x \leq 0
        \end{cases}

    .. seealso::

        **Original reference**: Godfrey, Luke B., and Michael S. Gashler.
        *A continuum among logarithmic, linear, and exponential functions,
        and its potential to improve generalization in neural networks.*
        2015 7th international joint conference on knowledge discovery,
        knowledge engineering and knowledge management (IC3K).
        Vol. 1. IEEE, 2015. DOI: `arXiv preprint arXiv:1602.01321.
        <https://arxiv.org/abs/1602.01321>`_.

        Jagtap, Ameya D., Kenji Kawaguchi, and George Em Karniadakis. *Adaptive
        activation functions accelerate convergence in deep and
        physics-informed neural networks*. Journal of
        Computational Physics 404 (2020): 109136. 
        DOI: `JCP 10.1016
        <https://doi.org/10.1016/j.jcp.2019.109136>`_.
    """

    def __init__(self, alpha=None, beta=None, gamma=None, fixed=None):
        super().__init__(alpha, beta, gamma, fixed)
        self._func = torch.nn.ELU()


class AdaptiveCELU(AdaptiveActivationFunctionInterface):
    r"""
    Adaptive trainable :class:`~torch.nn.CELU` activation function.

    Given the function :math:`\text{CELU}:\mathbb{R}^n\rightarrow\mathbb{R}^n`,
    the adaptive function
    :math:`\text{CELU}_{\text{adaptive}}:\mathbb{R}^n\rightarrow\mathbb{R}^n`
    is defined as:

    .. math::
        \text{CELU}_{\text{adaptive}}({x})=\alpha\,\text{CELU}(\beta{x}+\gamma),

    where :math:`\alpha,\,\beta,\,\gamma` are trainable parameters, and the
    CELU function is defined as:

    .. math::
        \text{CELU}(x) = \max(0,x) + \min(0, \alpha * (\exp(x) - 1))

    .. seealso::

        **Original reference**: Godfrey, Luke B., and Michael S. Gashler.
        *A continuum among logarithmic, linear, and exponential functions,
        and its potential to improve generalization in neural networks.*
        2015 7th international joint conference on knowledge discovery,
        knowledge engineering and knowledge management (IC3K).
        Vol. 1. IEEE, 2015. DOI: `arXiv preprint arXiv:1602.01321.
        <https://arxiv.org/abs/1602.01321>`_.

        Jagtap, Ameya D., Kenji Kawaguchi, and George Em Karniadakis. *Adaptive
        activation functions accelerate convergence in deep and
        physics-informed neural networks*. Journal of
        Computational Physics 404 (2020): 109136.
        DOI: `JCP 10.1016
        <https://doi.org/10.1016/j.jcp.2019.109136>`_.
    """

    def __init__(self, alpha=None, beta=None, gamma=None, fixed=None):
        super().__init__(alpha, beta, gamma, fixed)
        self._func = torch.nn.CELU()


class AdaptiveGELU(AdaptiveActivationFunctionInterface):
    r"""
    Adaptive trainable :class:`~torch.nn.GELU` activation function.

    Given the function :math:`\text{GELU}:\mathbb{R}^n\rightarrow\mathbb{R}^n`,
    the adaptive function
    :math:`\text{GELU}_{\text{adaptive}}:\mathbb{R}^n\rightarrow\mathbb{R}^n`
    is defined as:

    .. math::
        \text{GELU}_{\text{adaptive}}({x})=\alpha\,\text{GELU}(\beta{x}+\gamma),

    where :math:`\alpha,\,\beta,\,\gamma` are trainable parameters, and the
    GELU function is defined as:

    .. math::
        \text{GELU}(x)=0.5*x*(1+\text{Tanh}(\sqrt{2 / \pi}*(x+0.044715*x^3)))


    .. seealso::

        **Original reference**: Godfrey, Luke B., and Michael S. Gashler.
        *A continuum among logarithmic, linear, and exponential functions,
        and its potential to improve generalization in neural networks.*
        2015 7th international joint conference on knowledge discovery,
        knowledge engineering and knowledge management (IC3K).
        Vol. 1. IEEE, 2015. DOI: `arXiv preprint arXiv:1602.01321.
        <https://arxiv.org/abs/1602.01321>`_.

        Jagtap, Ameya D., Kenji Kawaguchi, and George Em Karniadakis. *Adaptive
        activation functions accelerate convergence in deep and
        physics-informed neural networks*. Journal of
        Computational Physics 404 (2020): 109136.
        DOI: `JCP 10.1016
        <https://doi.org/10.1016/j.jcp.2019.109136>`_.
    """

    def __init__(self, alpha=None, beta=None, gamma=None, fixed=None):
        super().__init__(alpha, beta, gamma, fixed)
        self._func = torch.nn.GELU()


class AdaptiveSoftmin(AdaptiveActivationFunctionInterface):
    r"""
    Adaptive trainable :class:`~torch.nn.Softmin` activation function.

    Given the function
    :math:`\text{Softmin}:\mathbb{R}^n\rightarrow\mathbb{R}^n`,
    the adaptive function
    :math:`\text{Softmin}_{\text{adaptive}}:\mathbb{R}^n\rightarrow\mathbb{R}^n`
    is defined as:

    .. math::
        \text{Softmin}_{\text{adaptive}}({x})=\alpha\,
        \text{Softmin}(\beta{x}+\gamma),

    where :math:`\alpha,\,\beta,\,\gamma` are trainable parameters, and the
    Softmin function is defined as:

    .. math::
        \text{Softmin}(x_{i}) = \frac{\exp(-x_i)}{\sum_j \exp(-x_j)}

    .. seealso::

        **Original reference**: Godfrey, Luke B., and Michael S. Gashler.
        *A continuum among logarithmic, linear, and exponential functions,
        and its potential to improve generalization in neural networks.*
        2015 7th international joint conference on knowledge discovery,
        knowledge engineering and knowledge management (IC3K).
        Vol. 1. IEEE, 2015. DOI: `arXiv preprint arXiv:1602.01321.
        <https://arxiv.org/abs/1602.01321>`_.

        Jagtap, Ameya D., Kenji Kawaguchi, and George Em Karniadakis. *Adaptive
        activation functions accelerate convergence in deep and
        physics-informed neural networks*. Journal of
        Computational Physics 404 (2020): 109136.
        DOI: `JCP 10.1016
        <https://doi.org/10.1016/j.jcp.2019.109136>`_.
    """

    def __init__(self, alpha=None, beta=None, gamma=None, fixed=None):
        super().__init__(alpha, beta, gamma, fixed)
        self._func = torch.nn.Softmin()


class AdaptiveSoftmax(AdaptiveActivationFunctionInterface):
    r"""
    Adaptive trainable :class:`~torch.nn.Softmax` activation function.

    Given the function
    :math:`\text{Softmax}:\mathbb{R}^n\rightarrow\mathbb{R}^n`,
    the adaptive function
    :math:`\text{Softmax}_{\text{adaptive}}:\mathbb{R}^n\rightarrow\mathbb{R}^n`
    is defined as:

    .. math::
        \text{Softmax}_{\text{adaptive}}({x})=\alpha\,
        \text{Softmax}(\beta{x}+\gamma),

    where :math:`\alpha,\,\beta,\,\gamma` are trainable parameters, and the
    Softmax function is defined as:

    .. math::
        \text{Softmax}(x_{i}) = \frac{\exp(x_i)}{\sum_j \exp(x_j)}

    .. seealso::

        **Original reference**: Godfrey, Luke B., and Michael S. Gashler.
        *A continuum among logarithmic, linear, and exponential functions,
        and its potential to improve generalization in neural networks.*
        2015 7th international joint conference on knowledge discovery,
        knowledge engineering and knowledge management (IC3K).
        Vol. 1. IEEE, 2015. DOI: `arXiv preprint arXiv:1602.01321.
        <https://arxiv.org/abs/1602.01321>`_.

        Jagtap, Ameya D., Kenji Kawaguchi, and George Em Karniadakis. *Adaptive
        activation functions accelerate convergence in deep and
        physics-informed neural networks*. Journal of
        Computational Physics 404 (2020): 109136.
        DOI: `JCP 10.1016
        <https://doi.org/10.1016/j.jcp.2019.109136>`_.
    """

    def __init__(self, alpha=None, beta=None, gamma=None, fixed=None):
        super().__init__(alpha, beta, gamma, fixed)
        self._func = torch.nn.Softmax()


class AdaptiveSIREN(AdaptiveActivationFunctionInterface):
    r"""
    Adaptive trainable :obj:`~torch.sin` function.

    Given the function :math:`\text{sin}:\mathbb{R}^n\rightarrow\mathbb{R}^n`,
    the adaptive function
    :math:`\text{sin}_{\text{adaptive}}:\mathbb{R}^n\rightarrow\mathbb{R}^n`
    is defined as:

    .. math::
        \text{sin}_{\text{adaptive}}({x}) = \alpha\,\text{sin}(\beta{x}+\gamma),

    where :math:`\alpha,\,\beta,\,\gamma` are trainable parameters.

    .. seealso::

        **Original reference**: Godfrey, Luke B., and Michael S. Gashler.
        *A continuum among logarithmic, linear, and exponential functions,
        and its potential to improve generalization in neural networks.*
        2015 7th international joint conference on knowledge discovery,
        knowledge engineering and knowledge management (IC3K).
        Vol. 1. IEEE, 2015. DOI: `arXiv preprint arXiv:1602.01321.
        <https://arxiv.org/abs/1602.01321>`_.

        Jagtap, Ameya D., Kenji Kawaguchi, and George Em Karniadakis. *Adaptive
        activation functions accelerate convergence in deep and
        physics-informed neural networks*. Journal of
        Computational Physics 404 (2020): 109136.
        DOI: `JCP 10.1016
        <https://doi.org/10.1016/j.jcp.2019.109136>`_.
    """

    def __init__(self, alpha=None, beta=None, gamma=None, fixed=None):
        super().__init__(alpha, beta, gamma, fixed)
        self._func = torch.sin


class AdaptiveExp(AdaptiveActivationFunctionInterface):
    r"""
    Adaptive trainable :obj:`~torch.exp` function.

    Given the function :math:`\text{exp}:\mathbb{R}^n\rightarrow\mathbb{R}^n`,
    the adaptive function
    :math:`\text{exp}_{\text{adaptive}}:\mathbb{R}^n\rightarrow\mathbb{R}^n`
    is defined as:

    .. math::
        \text{exp}_{\text{adaptive}}({x}) = \alpha\,\text{exp}(\beta{x}),

    where :math:`\alpha,\,\beta` are trainable parameters.

    .. seealso::

        **Original reference**: Godfrey, Luke B., and Michael S. Gashler.
        *A continuum among logarithmic, linear, and exponential functions,
        and its potential to improve generalization in neural networks.*
        2015 7th international joint conference on knowledge discovery,
        knowledge engineering and knowledge management (IC3K).
        Vol. 1. IEEE, 2015. DOI: `arXiv preprint arXiv:1602.01321.
        <https://arxiv.org/abs/1602.01321>`_.

        Jagtap, Ameya D., Kenji Kawaguchi, and George Em Karniadakis. *Adaptive
        activation functions accelerate convergence in deep and
        physics-informed neural networks*. Journal of
        Computational Physics 404 (2020): 109136.
        DOI: `JCP 10.1016
        <https://doi.org/10.1016/j.jcp.2019.109136>`_.
    """

    def __init__(self, alpha=None, beta=None, fixed=None):

        # only alpha, and beta parameters (gamma=0 fixed)
        if fixed is None:
            fixed = ["gamma"]
        else:
            check_consistency(fixed, str)
            fixed = list(fixed) + ["gamma"]

        # calling super
        super().__init__(alpha, beta, 0.0, fixed)
        self._func = torch.exp



"""Module for LabelTensor"""

from copy import copy, deepcopy
import torch
from torch import Tensor


class LabelTensor(torch.Tensor):
    """
    Extension of the :class:`torch.Tensor` class that includes labels for
    each dimension.
    """

    @staticmethod
    def __new__(cls, x, labels, *args, **kwargs):
        """
        Create a new instance of the :class:`~pina.label_tensor.LabelTensor`
        class.

        :param torch.Tensor x: :class:`torch.tensor` instance to be casted as a
            :class:`~pina.label_tensor.LabelTensor`.
        :param labels: Labels to assign to the tensor.
        :type labels: str | list[str] | dict
        :return: The instance of the :class:`~pina.label_tensor.LabelTensor`
            class.
        :rtype: LabelTensor
        """

        if isinstance(x, LabelTensor):
            return x
        return super().__new__(cls, x, *args, **kwargs)

    @property
    def tensor(self):
        """
        Returns the tensor part of the :class:`~pina.label_tensor.LabelTensor`
        object.

        :return: Tensor part of the :class:`~pina.label_tensor.LabelTensor`.
        :rtype: torch.Tensor
        """

        return self.as_subclass(Tensor)

    def __init__(self, x, labels):
        """
        Initialize the :class:`~pina.label_tensor.LabelTensor` instance, by
        checking the consistency of the labels and the tensor. Specifically, the
        labels must match the following conditions:

        - At each dimension, the number of labels must match the size of the \
            dimension.
        - At each dimension, the labels must be unique.

        The labels can be passed in the following formats:

        :Example:
            >>> from pina import LabelTensor
            >>> tensor = LabelTensor(
            >>>     torch.rand((2000, 3)),
            ...     {1: {"name": "space", "dof": ['a', 'b', 'c']}})
            >>> tensor = LabelTensor(
            >>>     torch.rand((2000, 3)),
            ...     ["a", "b", "c"])
        
        The keys of the dictionary are the dimension indices, and the values are
        dictionaries containing the labels and the name of the dimension. If 
        the labels are passed as a list, these are assigned to the last 
        dimension.

        :param torch.Tensor x: The tensor to be casted as a
            :class:`~pina.label_tensor.LabelTensor`.
        :param labels: Labels to assign to the tensor.
        :type labels: str | list[str] | dict
        :raises ValueError: If the labels are not consistent with the tensor.
        """
        super().__init__()
        if labels is not None:
            self.labels = labels
        else:
            self._labels = {}

    @property
    def full_labels(self):
        """
        Returns the full labels of the tensor, even for the dimensions that are
        not labeled.

        :return: The full labels of the tensor
        :rtype: dict
        """
        to_return_dict = {}
        shape_tensor = self.shape
        for i, value in enumerate(shape_tensor):
            if i in self._labels:
                to_return_dict[i] = self._labels[i]
            else:
                to_return_dict[i] = {"dof": range(value), "name": i}
        return to_return_dict

    @property
    def stored_labels(self):
        """
        Returns the labels stored inside the instance.

        :return: The labels stored inside the instance.
        :rtype: dict
        """
        return self._labels

    @property
    def labels(self):
        """
        Returns the labels of the last dimension of the instance.

        :return: labels of last dimension
        :rtype: list
        """
        if self.ndim - 1 in self._labels:
            return self._labels[self.ndim - 1]["dof"]
        return None

    @labels.setter
    def labels(self, labels):
        """
        Set labels stored insider the instance by checking the type of the
        input labels and handling it accordingly. The following types are
        accepted:

        - **list**: The list of labels is assigned to the last dimension.
        - **dict**: The dictionary of labels is assigned to the tensor.
        - **str**: The string is assigned to the last dimension.

        :param labels: Labels to assign to the class variable _labels.
        :type labels: str | list[str] | dict
        """

        if not hasattr(self, "_labels"):
            self._labels = {}
        if isinstance(labels, dict):
            self._init_labels_from_dict(labels)
        elif isinstance(labels, (list, range)):
            self._init_labels_from_list(labels)
        elif isinstance(labels, str):
            labels = [labels]
            self._init_labels_from_list(labels)
        else:
            raise ValueError("labels must be list, dict or string.")

    def _init_labels_from_dict(self, labels):
        """
        Store the internal label representation according to the values
        passed as input.

        :param dict labels: The label(s) to update.
        :raises ValueError: If the dof list contains duplicates or the number of
                            dof does not match the tensor shape.
        """

        tensor_shape = self.shape

        def validate_dof(dof_list, dim_size):
            """Validate the 'dof' list for uniqueness and size."""
            if len(dof_list) != len(set(dof_list)):
                raise ValueError("dof must be unique")
            if len(dof_list) != dim_size:
                raise ValueError(
                    f"Number of dof ({len(dof_list)}) does not match "
                    f"tensor shape ({dim_size})"
                )

        for dim, label in labels.items():
            if isinstance(label, dict):
                if "name" not in label:
                    label["name"] = dim
                if "dof" not in label:
                    label["dof"] = range(tensor_shape[dim])
                if "dof" in label and "name" in label:
                    dof = label["dof"]
                    dof_list = dof if isinstance(dof, (list, range)) else [dof]
                    if not isinstance(dof_list, (list, range)):
                        raise ValueError(
                            f"'dof' should be a list or range, not"
                            f" {type(dof_list)}"
                        )
                    validate_dof(dof_list, tensor_shape[dim])
                else:
                    raise ValueError(
                        "Labels dictionary must contain either "
                        " both 'name' and 'dof' keys"
                    )
            else:
                raise ValueError(
                    f"Invalid label format for {dim}: Expected "
                    f"list or dictionary, got {type(label)}"
                )

            # Assign validated label data to internal labels
            self._labels[dim] = label

    def _init_labels_from_list(self, labels):
        """
        Given a list of dof, this method update the internal label
        representation by assigning the dof to the last dimension.

        :param labels: The label(s) to update.
        :type labels: list
        """

        # Create a dict with labels
        last_dim_labels = {
            self.ndim - 1: {"dof": labels, "name": self.ndim - 1}
        }
        self._init_labels_from_dict(last_dim_labels)

    def extract(self, labels_to_extract):
        """
        Extract the subset of the original tensor by returning all the positions
        corresponding to the passed ``label_to_extract``. If
        ``label_to_extract`` is a dictionary, the keys are the dimension names
        and the values are the labels to extract. If a single label or a list
        of labels is passed, the last dimension is considered.

        :Example:
            >>> from pina import LabelTensor
            >>> labels = {1: {'dof': ["a", "b", "c"], 'name': 'space'}}
            >>> tensor = LabelTensor(torch.rand((2000, 3)), labels)
            >>> tensor.extract("a")
            >>> tensor.extract(["a", "b"])
            >>> tensor.extract({"space": ["a", "b"]})

        :param labels_to_extract: The label(s) to extract.
        :type labels_to_extract: str | list[str] | tuple[str] | dict
        :return: The extracted tensor with the updated labels.
        :rtype: LabelTensor

        :raises TypeError: Labels are not ``str``, ``list[str]`` or ``dict``
            properly setted.
        :raises ValueError: Label to extract is not in the labels ``list``.
        """

        def get_label_indices(dim_labels, labels_te):
            if isinstance(labels_te, (int, str)):
                labels_te = [labels_te]
            return (
                [dim_labels.index(label) for label in labels_te]
                if len(labels_te) > 1
                else slice(
                    dim_labels.index(labels_te[0]),
                    dim_labels.index(labels_te[0]) + 1,
                )
            )

        # Ensure labels_to_extract is a list or dict
        if isinstance(labels_to_extract, (str, int)):
            labels_to_extract = [labels_to_extract]

        labels = copy(self._labels)

        # Get the dimension names and the respective dimension index
        dim_names = {labels[dim]["name"]: dim for dim in labels}
        ndim = super().ndim
        tensor = self.tensor.as_subclass(torch.Tensor)

        # Convert list/tuple to a dict for the last dimension if applicable
        if isinstance(labels_to_extract, (list, tuple)):
            last_dim = ndim - 1
            dim_name = labels[last_dim]["name"]
            labels_to_extract = {dim_name: list(labels_to_extract)}

        # Validate the labels_to_extract type
        if not isinstance(labels_to_extract, dict):
            raise ValueError(
                "labels_to_extract must be a string, list, or dictionary."
            )

        # Perform the extraction for each specified dimension
        for dim_name, labels_te in labels_to_extract.items():
            if dim_name not in dim_names:
                raise ValueError(
                    f"Cannot extract labels for dimension '{dim_name}' as it is"
                    f" not present in the original labels."
                )

            idx_dim = dim_names[dim_name]
            dim_labels = labels[idx_dim]["dof"]
            indices = get_label_indices(dim_labels, labels_te)

            extractor = [slice(None)] * ndim
            extractor[idx_dim] = indices
            tensor = tensor[tuple(extractor)]

            labels[idx_dim] = {"dof": labels_te, "name": dim_name}

        return LabelTensor(tensor, labels)

    def __str__(self):
        """
        The string representation of the
        :class:`~pina.label_tensor.LabelTensor`.

        :return: String representation of the
            :class:`~pina.label_tensor.LabelTensor` instance.
        :rtype: str
        """

        s = ""
        for key, value in self._labels.items():
            s += f"{key}: {value}\n"
        s += "\n"
        s += self.tensor.__str__()
        return s

    @staticmethod
    def cat(tensors, dim=0):
        """
        Concatenate a list of tensors along a specified dimension. For more
        details, see :meth:`torch.cat`.

        :param list[LabelTensor] tensors:
            :class:`~pina.label_tensor.LabelTensor` instances to concatenate
        :param int dim: Dimensions on which you want to perform the operation
            (default is 0)
        :return: A new :class:`LabelTensor` instance obtained by concatenating
            the input instances.

        :rtype: LabelTensor
        :raises ValueError: either number dof or dimensions names differ.
        """

        if not tensors:
            return []  # Handle empty list
        if len(tensors) == 1:
            return tensors[0]  # Return single tensor as-is

            # Perform concatenation
        cat_tensor = torch.cat(tensors, dim=dim)
        tensors_labels = [tensor.stored_labels for tensor in tensors]

        # Check label consistency across tensors, excluding the
        # concatenation dimension
        for key in tensors_labels[0]:
            if key != dim:
                if any(
                    tensors_labels[i][key] != tensors_labels[0][key]
                    for i in range(len(tensors_labels))
                ):
                    raise RuntimeError(
                        f"Tensors must have the same labels along all "
                        f"dimensions except {dim}."
                    )

        # Copy and update the 'dof' for the concatenation dimension
        cat_labels = {k: copy(v) for k, v in tensors_labels[0].items()}

        # Update labels if the concatenation dimension has labels
        if dim in tensors[0].stored_labels:
            if dim in cat_labels:
                cat_dofs = [label[dim]["dof"] for label in tensors_labels]
                cat_labels[dim]["dof"] = sum(cat_dofs, [])
        else:
            cat_labels = tensors[0].stored_labels

        # Assign updated labels to the concatenated tensor
        cat_tensor._labels = cat_labels
        return cat_tensor

    @staticmethod
    def stack(tensors):
        """
        Stacks a list of tensors along a new dimension. For more details, see
        :meth:`torch.stack`.

        :param list[LabelTensor] tensors: A list of tensors to stack.
            All tensors must have the same shape.
        :return: A new :class:`~pina.label_tensor.LabelTensor` instance obtained
            by stacking the input tensors.
        :rtype: LabelTensor
        """

        # Perform stacking in torch
        new_tensor = torch.stack(tensors)

        # Increase labels keys by 1
        labels = tensors[0]._labels
        labels = {key + 1: value for key, value in labels.items()}
        new_tensor._labels = labels
        return new_tensor

    def requires_grad_(self, mode=True):
        """
        Override the :meth:`~torch.Tensor.requires_grad_` method to handle
        the labels in the new tensor.
        For more details, see :meth:`~torch.Tensor.requires_grad_`.

        :param bool mode: A boolean value indicating whether the tensor should
            track gradients.If `True`, the tensor will track gradients;
            if `False`, it will not.
        :return: The :class:`~pina.label_tensor.LabelTensor` itself with the
            updated ``requires_grad`` state and retained labels.
        :rtype: LabelTensor
        """

        lt = super().requires_grad_(mode)
        lt._labels = self._labels
        return lt

    @property
    def dtype(self):
        """
        Give the ``dtype`` of the tensor. For more details, see
        :meth:`torch.dtype`.

        :return: The data type of the tensor.
        :rtype: torch.dtype
        """

        return super().dtype

    def to(self, *args, **kwargs):
        """
        Performs Tensor dtype and/or device conversion. For more details, see
        :meth:`torch.Tensor.to`.

        :return: A new :class:`~pina.label_tensor.LabelTensor` instance with the
            updated dtype and/or device and retained labels.
        :rtype: LabelTensor
        """

        lt = super().to(*args, **kwargs)
        lt._labels = self._labels
        return lt

    def clone(self, *args, **kwargs):
        """
        Clone the :class:`~pina.label_tensor.LabelTensor`. For more details, see
        :meth:`torch.Tensor.clone`.

        :return: A new :class:`~pina.label_tensor.LabelTensor` instance with the
            same data and labels but allocated in a different memory location.
        :rtype: LabelTensor
        """

        out = LabelTensor(
            super().clone(*args, **kwargs), deepcopy(self._labels)
        )
        return out

    def append(self, tensor, mode="std"):
        """
        Appends a given tensor to the current tensor along the last dimension.
        This method supports two types of appending operations:

        1. **Standard append** ("std"): Concatenates the input tensor with the \
            current tensor along the last dimension.
        2. **Cross append** ("cross"): Creates a cross-product of the current \
            tensor and the input tensor.

        :param tensor: The tensor to append to the current tensor.
        :type tensor: LabelTensor
        :param mode: The append mode to use. Defaults to ``st``.
        :type mode: str, optional
        :return: A new :class:`LabelTensor` instance obtained by appending the 
            input tensor.
        :rtype: LabelTensor

        :raises ValueError: If the mode is not "std" or "cross".
        """

        if mode == "std":
            # Call cat on last dimension
            new_label_tensor = LabelTensor.cat(
                [self, tensor], dim=self.ndim - 1
            )
            return new_label_tensor
        if mode == "cross":
            # Crete tensor and call cat on last dimension
            tensor1 = self
            tensor2 = tensor
            n1 = tensor1.shape[0]
            n2 = tensor2.shape[0]
            tensor1 = LabelTensor(tensor1.repeat(n2, 1), labels=tensor1.labels)
            tensor2 = LabelTensor(
                tensor2.repeat_interleave(n1, dim=0), labels=tensor2.labels
            )
            new_label_tensor = LabelTensor.cat(
                [tensor1, tensor2], dim=self.ndim - 1
            )
            return new_label_tensor
        raise ValueError('mode must be either "std" or "cross"')

    @staticmethod
    def vstack(tensors):
        """
        Stack tensors vertically. For more details, see :meth:`torch.vstack`.

        :param list of LabelTensor label_tensors: The
            :class:`~pina.label_tensor.LabelTensor` instances to stack. They
            need to have equal labels.
        :return: A new :class:`~pina.label_tensor.LabelTensor` instance obtained
            by stacking the input tensors vertically.
        :rtype: LabelTensor
        """

        return LabelTensor.cat(tensors, dim=0)

    # This method is used to update labels
    def _update_single_label(self, index, dim):
        """
        Update the labels of the tensor based on the index (or list of indices).

        :param index: Index of dof to retain.
        :type index: int | slice | list[int] | tuple[int] | torch.Tensor
        :param int dim: Dimension of the indexes in the original tensor.
        :return: The updated labels for the specified dimension.
        :rtype: list[int]
        :raises: ValueError: If the index type is not supported.
        """
        old_dof = self._labels[dim]["dof"]
        # Handle slicing
        if isinstance(index, slice):
            new_dof = old_dof[index]
        # Handle single integer index
        elif isinstance(index, int):
            new_dof = [old_dof[index]]
        # Handle lists or tensors
        elif isinstance(index, (list, torch.Tensor)):
            # Handle list of bools
            if isinstance(index, torch.Tensor) and index.dtype == torch.bool:
                index = index.nonzero().squeeze()
            new_dof = (
                [old_dof[i] for i in index]
                if isinstance(old_dof, list)
                else index
            )
        else:
            raise NotImplementedError(
                f"Unsupported index type: {type(index)}. Expected slice, int, "
                f"list, or torch.Tensor."
            )
        return new_dof

    def __getitem__(self, index):
        """ "
        Override the __getitem__ method to handle the labels of the
        :class:`~pina.label_tensor.LabelTensor` instance. It first performs
        __getitem__ operation on the :class:`torch.Tensor` part of the instance,
        then updates the labels based on the index.

        :param index: The index used to access the item
        :type index: int | str | tuple of int  | list ot int | torch.Tensor
        :return: A new :class:`~pina.label_tensor.LabelTensor` instance obtained
            `__getitem__` operation on :class:`torch.Tensor` part of the
            instance, with the updated labels.
        :rtype: LabelTensor

        :raises KeyError: If an invalid label index is provided.
        :raises IndexError: If an invalid index is accessed in the tensor.
        """

        # Handle string index
        if isinstance(index, str) or (
            isinstance(index, (tuple, list))
            and all(isinstance(i, str) for i in index)
        ):
            return self.extract(index)

        # Retrieve selected tensor and labels
        selected_tensor = super().__getitem__(index)
        if not hasattr(self, "_labels"):
            return selected_tensor

        original_labels = self._labels
        updated_labels = copy(original_labels)

        # Ensure the index is iterable
        if not isinstance(index, tuple):
            index = [index]

        # Update labels based on the index
        offset = 0
        removed = 0
        for dim, idx in enumerate(index):
            if dim in original_labels:
                if isinstance(idx, int):
                    # Compute the working dimension considering the removed
                    # dimensions due to int index on a non labled dimension
                    dim_ = dim - removed
                    selected_tensor = selected_tensor.unsqueeze(dim_)
                if idx != slice(None):
                    # Update the labels for the selected dimension
                    updated_labels[offset] = {
                        "dof": self._update_single_label(idx, dim),
                        "name": original_labels[dim]["name"],
                    }
            else:
                # Adjust label keys if dimension is reduced (case of integer
                # index on a non-labeled dimension)
                if isinstance(idx, int):
                    updated_labels = {
                        key - 1 if key > dim else key: value
                        for key, value in updated_labels.items()
                    }
                    removed += 1
                    continue
            offset += 1

        # Update the selected tensor's labels
        selected_tensor._labels = updated_labels
        return selected_tensor

    def sort_labels(self, dim=None):
        """
        Sort the labels along the specified dimension and apply. It applies the
        same sorting to the tensor part of the instance.

        :param int dim: The dimension along which to sort the labels.
            If ``None``, the last dimension is used.
        :return: A new tensor with sorted labels along the specified dimension.
        :rtype: LabelTensor
        """

        def arg_sort(lst):
            return sorted(range(len(lst)), key=lambda x: lst[x])

        if dim is None:
            dim = self.ndim - 1
        if self.shape[dim] == 1:
            return self
        labels = self.stored_labels[dim]["dof"]
        sorted_index = arg_sort(labels)
        # Define an indexer to sort the tensor along the specified dimension
        indexer = [slice(None)] * self.ndim
        # Assigned the sorted index to the specified dimension
        indexer[dim] = sorted_index
        return self[tuple(indexer)]

    def __deepcopy__(self, memo):
        """
        Creates a deep copy of the object. For more details, see
        :meth:`copy.deepcopy`.

        :param memo: LabelTensor object to be copied.
        :type memo: LabelTensor
        :return: A deep copy of the original LabelTensor object.
        :rtype: LabelTensor
        """

        cls = self.__class__
        result = cls(deepcopy(self.tensor), deepcopy(self.stored_labels))
        return result

    def permute(self, *dims):
        """
        Permutes the dimensions of the tensor and the associated labels
        accordingly. For more details, see :meth:`torch.Tensor.permute`.

        :param dims: The dimensions to permute the tensor to.
        :type dims: tuple[int] | list[int]
        :return: A new object with permuted dimensions and reordered labels.
        :rtype: LabelTensor
        """
        # Call the base class permute method
        tensor = super().permute(*dims)

        # Update lables
        labels = self._labels
        keys_list = list(*dims)
        labels = {keys_list.index(k): v for k, v in labels.items()}

        # Assign labels to the new tensor
        tensor._labels = labels
        return tensor

    def detach(self):
        """
        Detaches the tensor from the computation graph and retains the stored
        labels. For more details, see :meth:`torch.Tensor.detach`.

        :return: A new tensor detached from the computation graph.
        :rtype: LabelTensor
        """

        lt = super().detach()

        # Copy the labels to the new tensor only if present
        if hasattr(self, "_labels"):
            lt._labels = self.stored_labels
        return lt

    @staticmethod
    def summation(tensors):
        """
        Computes the summation of a list of
        :class:`~pina.label_tensor.LabelTensor` instances.


        :param list[LabelTensor] tensors: A list of tensors to sum. All
            tensors must have the same shape and labels.
        :return: A new `LabelTensor` containing the element-wise sum of the
                 input tensors.
        :rtype: LabelTensor

        :raises ValueError: If the input `tensors` list is empty.
        :raises RuntimeError: If the tensors have different shapes and/or
                              mismatched labels.
        """

        if not tensors:
            raise ValueError("The tensors list must not be empty.")

        if len(tensors) == 1:
            return tensors[0]

        # Initialize result tensor and labels
        data = torch.zeros_like(tensors[0].tensor).to(tensors[0].device)
        last_dim_labels = []

        # Accumulate tensors
        for tensor in tensors:
            data += tensor.tensor
            last_dim_labels.append(tensor.labels)

        # Construct last dimension labels
        last_dim_labels = ["+".join(items) for items in zip(*last_dim_labels)]

        # Update the labels for the resulting tensor
        labels = {k: copy(v) for k, v in tensors[0].stored_labels.items()}
        labels[tensors[0].ndim - 1] = {
            "dof": last_dim_labels,
            "name": tensors[0].name,
        }

        return LabelTensor(data, labels)

    def reshape(self, *shape):
        """
        Override the reshape method to update the labels of the tensor.
        For more details, see :meth:`torch.Tensor.reshape`.

        :param tuple of int shape: The new shape of the tensor.
        :return: A new :class:`~pina.label_tensor.LabelTensor` instance with the
            updated shape and labels.
        :rtype: LabelTensor
        """

        # As for now the reshape method is used only in the context of the
        # dataset, the labels are not
        tensor = super().reshape(*shape)
        if not hasattr(self, "_labels") or shape != (-1, *self.shape[2:]):
            return tensor
        tensor.labels = self.labels
        return tensor



"""Module to create and handle domains."""

__all__ = [
    "DomainInterface",
    "CartesianDomain",
    "EllipsoidDomain",
    "Union",
    "Intersection",
    "Exclusion",
    "Difference",
    "OperationInterface",
    "SimplexDomain",
]

from .domain_interface import DomainInterface
from .cartesian import CartesianDomain
from .ellipsoid import EllipsoidDomain
from .exclusion_domain import Exclusion
from .intersection_domain import Intersection
from .union_domain import Union
from .difference_domain import Difference
from .operation_interface import OperationInterface
from .simplex import SimplexDomain



"""Module for the Gradient PINN solver."""

import torch

from .pinn import PINN
from ...operator import grad
from ...problem import SpatialProblem


class GradientPINN(PINN):
    r"""
    Gradient Physics-Informed Neural Network (GradientPINN) solver class.
    This class implements the Gradient Physics-Informed Neural Network solver,
    using a user specified ``model`` to solve a specific ``problem``.
    It can be used to solve both forward and inverse problems.

    The Gradient Physics-Informed Neural Network solver aims to find the
    solution :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m` of a differential
    problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}

    minimizing the loss function;

    .. math::
        \mathcal{L}_{\rm{problem}} =& \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathcal{A}[\mathbf{u}](\mathbf{x}_i)) +
        \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathcal{B}[\mathbf{u}](\mathbf{x}_i)) + 
        &\frac{1}{N}\sum_{i=1}^N
        \nabla_{\mathbf{x}}\mathcal{L}(\mathcal{A}[\mathbf{u}](\mathbf{x}_i)) +
        \frac{1}{N}\sum_{i=1}^N
        \nabla_{\mathbf{x}}\mathcal{L}(\mathcal{B}[\mathbf{u}](\mathbf{x}_i))


    where :math:`\mathcal{L}` is a specific loss function, typically the MSE:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.

    .. seealso::

        **Original reference**: Yu, Jeremy, et al.
        *Gradient-enhanced physics-informed neural networks for forward and
        inverse PDE problems.*
        Computer Methods in Applied Mechanics and Engineering 393 (2022):114823.
        DOI: `10.1016 <https://doi.org/10.1016/j.cma.2022.114823>`_.

    .. note::
        This class is only compatible with problems that inherit from  the
        :class:`~pina.problem.spatial_problem.SpatialProblem` class. 
    """

    def __init__(
        self,
        problem,
        model,
        optimizer=None,
        scheduler=None,
        weighting=None,
        loss=None,
    ):
        """
        Initialization of the :class:`GradientPINN` class.

        :param AbstractProblem problem: The problem to be solved.
            It must inherit from at least
            :class:`~pina.problem.spatial_problem.SpatialProblem` to compute the
            gradient of the loss.
        :param torch.nn.Module model: The neural network model to be used.
        :param Optimizer optimizer: The optimizer to be used.
            If ``None``, the :class:`torch.optim.Adam` optimizer is used.
            Default is ``None``.
        :param Scheduler scheduler: Learning rate scheduler.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param WeightingInterface weighting: The weighting schema to be used.
            If ``None``, no weighting schema is used. Default is ``None``.
        :param torch.nn.Module loss: The loss function to be minimized.
            If ``None``, the :class:`torch.nn.MSELoss` loss is used.
            Default is `None`.
        :raises ValueError: If the problem is not a SpatialProblem.
        """
        super().__init__(
            model=model,
            problem=problem,
            optimizer=optimizer,
            scheduler=scheduler,
            weighting=weighting,
            loss=loss,
        )

        if not isinstance(self.problem, SpatialProblem):
            raise ValueError(
                "Gradient PINN computes the gradient of the "
                "PINN loss with respect to the spatial "
                "coordinates, thus the PINA problem must be "
                "a SpatialProblem."
            )

    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the physics-informed solver based on the
        provided samples and equation.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation.
        :return: The computed physics loss.
        :rtype: LabelTensor
        """
        # classical PINN loss
        residual = self.compute_residual(samples=samples, equation=equation)
        loss_value = self._loss_fn(
            torch.zeros_like(residual, requires_grad=True), residual
        )

        # gradient PINN loss
        loss_value = loss_value.reshape(-1, 1)
        loss_value.labels = ["__loss"]
        loss_grad = grad(loss_value, samples, d=self.problem.spatial_variables)
        g_loss_phys = self._loss_fn(
            torch.zeros_like(loss_grad, requires_grad=True), loss_grad
        )
        return loss_value + g_loss_phys



"""Module for the Adaptive Function interface."""

from abc import ABCMeta
import torch
from ..utils import check_consistency, is_function


class AdaptiveActivationFunctionInterface(torch.nn.Module, metaclass=ABCMeta):
    r"""
    The :class:`AdaptiveActivationFunctionInterface`
    class makes a :class:`torch.nn.Module` activation function into an adaptive
    trainable activation function. If one wants to create an adpative activation
    function, this class must be use as base class.

    Given a function :math:`f:\mathbb{R}^n\rightarrow\mathbb{R}^m`, the adaptive
    function :math:`f_{\text{adaptive}}:\mathbb{R}^n\rightarrow\mathbb{R}^m`
    is defined as:

    .. math::
        f_{\text{adaptive}}(\mathbf{x}) = \alpha\,f(\beta\mathbf{x}+\gamma),

    where :math:`\alpha,\,\beta,\,\gamma` are trainable parameters.

    .. seealso::

        **Original reference**: Godfrey, Luke B., and Michael S. Gashler.
        *A continuum among logarithmic, linear, and exponential functions,
        and its potential to improve generalization in neural networks.*
        2015 7th international joint conference on knowledge discovery,
        knowledge engineering and knowledge management (IC3K).
        Vol. 1. IEEE, 2015. DOI: `arXiv preprint arXiv:1602.01321.
        <https://arxiv.org/abs/1602.01321>`_.

        Jagtap, Ameya D., Kenji Kawaguchi, and George Em Karniadakis. *Adaptive
        activation functions accelerate convergence in deep and
        physics-informed neural networks*. Journal of
        Computational Physics 404 (2020): 109136.
        DOI: `JCP 10.1016
        <https://doi.org/10.1016/j.jcp.2019.109136>`_.
    """

    def __init__(self, alpha=None, beta=None, gamma=None, fixed=None):
        """
        Initializes the Adaptive Function.

        :param float | complex alpha: Scaling parameter alpha.
            Defaults to ``None``. When ``None`` is passed,
            the variable is initialized to 1.
        :param float | complex beta: Scaling parameter beta.
            Defaults to ``None``. When ``None`` is passed,
            the variable is initialized to 1.
        :param float | complex gamma: Shifting parameter gamma.
            Defaults to ``None``. When ``None`` is passed,
            the variable is initialized to 1.
        :param list fixed: List of parameters to fix during training,
            i.e. not optimized (``requires_grad`` set to ``False``).
            Options are ``alpha``, ``beta``, ``gamma``. Defaults to None.
        """
        super().__init__()

        # see if there are fixed variables
        if fixed is not None:
            check_consistency(fixed, str)
            if not all(key in ["alpha", "beta", "gamma"] for key in fixed):
                raise TypeError(
                    "Fixed keys must be in [`alpha`, `beta`, `gamma`]."
                )

        # initialize alpha, beta, gamma if they are None
        if alpha is None:
            alpha = 1.0
        if beta is None:
            beta = 1.0
        if gamma is None:
            gamma = 0.0

        # checking consistency
        check_consistency(alpha, (float, complex))
        check_consistency(beta, (float, complex))
        check_consistency(gamma, (float, complex))

        # registering as tensors
        alpha = torch.tensor(alpha, requires_grad=False)
        beta = torch.tensor(beta, requires_grad=False)
        gamma = torch.tensor(gamma, requires_grad=False)

        # setting not fixed variables as torch.nn.Parameter with gradient
        # registering the buffer for the one which are fixed, buffers by
        # default are saved alongside trainable parameters
        if "alpha" not in (fixed or []):
            self._alpha = torch.nn.Parameter(alpha, requires_grad=True)
        else:
            self.register_buffer("alpha", alpha)

        if "beta" not in (fixed or []):
            self._beta = torch.nn.Parameter(beta, requires_grad=True)
        else:
            self.register_buffer("beta", beta)

        if "gamma" not in (fixed or []):
            self._gamma = torch.nn.Parameter(gamma, requires_grad=True)
        else:
            self.register_buffer("gamma", gamma)

    def forward(self, x):
        """
        Define the computation performed at every call.
        The function to the input elementwise.

        :param x: The input tensor to evaluate the activation function.
        :type x: torch.Tensor | LabelTensor
        """
        return self.alpha * (self._func(self.beta * x + self.gamma))

    @property
    def alpha(self):
        """
        The alpha variable.
        """
        return self._alpha

    @property
    def beta(self):
        """
        The beta variable.
        """
        return self._beta

    @property
    def gamma(self):
        """
        The gamma variable.
        """
        return self._gamma

    @property
    def func(self):
        """
        The callable activation function.
        """
        return self._func

    @func.setter
    def func(self, value):
        """
        Set the activation function.
        """
        if not is_function(value):
            raise TypeError("The function must be callable.")
        self._func = value
        return self._func



"""Formulation of the advection problem."""

import torch
from ... import Condition
from ...operator import grad
from ...equation import Equation
from ...domain import CartesianDomain
from ...utils import check_consistency
from ...problem import SpatialProblem, TimeDependentProblem


class AdvectionEquation(Equation):
    """
    Implementation of the advection equation.
    """

    def __init__(self, c):
        """
        Initialization of the :class:`AdvectionEquation`.

        :param c: The advection velocity parameter.
        :type c: float | int
        """
        self.c = c
        check_consistency(self.c, (float, int))

        def equation(input_, output_):
            """
            Implementation of the advection equation.

            :param LabelTensor input_: Input data of the problem.
            :param LabelTensor output_: Output data of the problem.
            :return: The residual of the advection equation.
            :rtype: LabelTensor
            """
            u_x = grad(output_, input_, components=["u"], d=["x"])
            u_t = grad(output_, input_, components=["u"], d=["t"])
            return u_t + self.c * u_x

        super().__init__(equation)


def initial_condition(input_, output_):
    """
    Implementation of the initial condition.

    :param LabelTensor input_: Input data of the problem.
    :param LabelTensor output_: Output data of the problem.
    :return: The residual of the initial condition.
    :rtype: LabelTensor
    """
    return output_ - torch.sin(input_.extract("x"))


class AdvectionProblem(SpatialProblem, TimeDependentProblem):
    r"""
    Implementation of the advection problem in the spatial interval
    :math:`[0, 2 \pi]` and temporal interval :math:`[0, 1]`.

    .. seealso::

        **Original reference**: Wang, Sifan, et al. *An expert's guide to
        training physics-informed neural networks*.
        arXiv preprint arXiv:2308.08468 (2023).
        DOI: `arXiv:2308.08468  <https://arxiv.org/abs/2308.08468>`_.

    :Example:
        >>> problem = AdvectionProblem(c=1.0)
    """

    output_variables = ["u"]
    spatial_domain = CartesianDomain({"x": [0, 2 * torch.pi]})
    temporal_domain = CartesianDomain({"t": [0, 1]})

    domains = {
        "D": CartesianDomain({"x": [0, 2 * torch.pi], "t": [0, 1]}),
        "t0": CartesianDomain({"x": [0, 2 * torch.pi], "t": 0.0}),
    }

    conditions = {
        "t0": Condition(domain="t0", equation=Equation(initial_condition)),
    }

    def __init__(self, c=1.0):
        """
        Initialization of the :class:`AdvectionProblem`.

        :param c: The advection velocity parameter.
        :type c: float | int
        """
        super().__init__()

        self.c = c
        check_consistency(self.c, (float, int))

        self.conditions["D"] = Condition(
            domain="D", equation=AdvectionEquation(self.c)
        )

    def solution(self, pts):
        """
        Implementation of the analytical solution of the advection problem.

        :param LabelTensor pts: Points where the solution is evaluated.
        :return: The analytical solution of the advection problem.
        :rtype: LabelTensor
        """
        sol = torch.sin(pts.extract("x") - self.c * pts.extract("t"))
        sol.labels = self.output_variables
        return sol



"""Module for the Condition class."""

import warnings
from .data_condition import DataCondition
from .domain_equation_condition import DomainEquationCondition
from .input_equation_condition import InputEquationCondition
from .input_target_condition import InputTargetCondition
from ..utils import custom_warning_format

# Set the custom format for warnings
warnings.formatwarning = custom_warning_format
warnings.filterwarnings("always", category=DeprecationWarning)


def warning_function(new, old):
    """Handle the deprecation warning.

    :param new: Object to use instead of the old one.
    :type new: str
    :param old: Object to deprecate.
    :type old: str
    """
    warnings.warn(
        f"'{old}' is deprecated and will be removed "
        f"in future versions. Please use '{new}' instead.",
        DeprecationWarning,
    )


class Condition:
    """
    Represents constraints (such as physical equations, boundary conditions,
    etc.) that must be satisfied in a given problem. Condition objects are used
    to formulate the PINA
    :class:`~pina.problem.abstract_problem.AbstractProblem` object.

    There are different types of conditions:

    - :class:`~pina.condition.input_target_condition.InputTargetCondition`:
      Defined by specifying both the input and the target of the condition. In
      this case, the model is trained to produce the target given the input. The
      input and output   data must be one of the :class:`torch.Tensor`,
      :class:`~pina.label_tensor.LabelTensor`,
      :class:`~torch_geometric.data.Data`, or :class:`~pina.graph.Graph`.
      Different implementations exist depending on the type of input and target.
      For more details, see
      :class:`~pina.condition.input_target_condition.InputTargetCondition`.

    - :class:`~pina.condition.domain_equation_condition.DomainEquationCondition`
      : Defined by specifying both the domain and the equation of the condition.
      Here, the model is trained to minimize the equation residual by evaluating
      it at sampled points within the domain.

    - :class:`~pina.condition.input_equation_condition.InputEquationCondition`:
      Defined by specifying the input and the equation of the condition. In this
      case, the model is trained to minimize the equation residual by evaluating
      it at the provided input. The input must be either a
      :class:`~pina.label_tensor.LabelTensor` or a :class:`~pina.graph.Graph`.
      Different implementations exist depending on the type of input. For more
      details, see
      :class:`~pina.condition.input_equation_condition.InputEquationCondition`.

    - :class:`~pina.condition.data_condition.DataCondition`:
      Defined by specifying only the input. In this case, the model is trained
      with an unsupervised custom loss while using the provided data during
      training. The input data must be one of :class:`torch.Tensor`,
      :class:`~pina.label_tensor.LabelTensor`,
      :class:`~torch_geometric.data.Data`, or :class:`~pina.graph.Graph`.
      Additionally, conditional variables can be provided when the model
      depends on extra parameters. These conditional variables must be either
      :class:`torch.Tensor` or :class:`~pina.label_tensor.LabelTensor`.
      Different implementations exist depending on the type of input.
      For more details, see
      :class:`~pina.condition.data_condition.DataCondition`.

    :Example:

    >>> from pina import Condition
    >>> condition = Condition(
    ...     input=input,
    ...     target=target
    ... )
    >>> condition = Condition(
    ...     domain=location,
    ...     equation=equation
    ... )
    >>> condition = Condition(
    ...     input=input,
    ...     equation=equation
    ... )
    >>> condition = Condition(
    ...     input=data,
    ...     conditional_variables=conditional_variables
    ... )

    """

    __slots__ = list(
        set(
            InputTargetCondition.__slots__
            + InputEquationCondition.__slots__
            + DomainEquationCondition.__slots__
            + DataCondition.__slots__
        )
    )

    def __new__(cls, *args, **kwargs):
        """
        Instantiate the appropriate Condition object based on the keyword
        arguments passed.

        :raises ValueError: If no keyword arguments are passed.
        :raises ValueError: If the keyword arguments are invalid.
        :return: The appropriate Condition object.
        :rtype: ConditionInterface
        """

        if len(args) != 0:
            raise ValueError(
                "Condition takes only the following keyword "
                f"arguments: {Condition.__slots__}."
            )

        # back-compatibility 0.1
        keys = list(kwargs.keys())
        if "location" in keys:
            kwargs["domain"] = kwargs.pop("location")
            warning_function(new="domain", old="location")

        if "input_points" in keys:
            kwargs["input"] = kwargs.pop("input_points")
            warning_function(new="input", old="input_points")

        if "output_points" in keys:
            kwargs["target"] = kwargs.pop("output_points")
            warning_function(new="target", old="output_points")

        sorted_keys = sorted(kwargs.keys())
        if sorted_keys == sorted(InputTargetCondition.__slots__):
            return InputTargetCondition(**kwargs)
        if sorted_keys == sorted(InputEquationCondition.__slots__):
            return InputEquationCondition(**kwargs)
        if sorted_keys == sorted(DomainEquationCondition.__slots__):
            return DomainEquationCondition(**kwargs)
        if (
            sorted_keys == sorted(DataCondition.__slots__)
            or sorted_keys[0] == DataCondition.__slots__[0]
        ):
            return DataCondition(**kwargs)

        raise ValueError(f"Invalid keyword arguments {kwargs.keys()}.")



"""Module for PINA Conditions classes."""

__all__ = [
    "Condition",
    "ConditionInterface",
    "DomainEquationCondition",
    "InputTargetCondition",
    "TensorInputTensorTargetCondition",
    "TensorInputGraphTargetCondition",
    "GraphInputTensorTargetCondition",
    "GraphInputGraphTargetCondition",
    "InputEquationCondition",
    "InputTensorEquationCondition",
    "InputGraphEquationCondition",
    "DataCondition",
    "GraphDataCondition",
    "TensorDataCondition",
]

from .condition_interface import ConditionInterface
from .condition import Condition
from .domain_equation_condition import DomainEquationCondition
from .input_target_condition import (
    InputTargetCondition,
    TensorInputTensorTargetCondition,
    TensorInputGraphTargetCondition,
    GraphInputTensorTargetCondition,
    GraphInputGraphTargetCondition,
)
from .input_equation_condition import (
    InputEquationCondition,
    InputTensorEquationCondition,
    InputGraphEquationCondition,
)
from .data_condition import (
    DataCondition,
    GraphDataCondition,
    TensorDataCondition,
)



"""
This module contains the PinaDataModule class, which extends the
LightningDataModule class to allow proper creation and management of
different types of Datasets defined in PINA.
"""

import warnings
from lightning.pytorch import LightningDataModule
import torch
from torch_geometric.data import Data
from torch.utils.data import DataLoader, SequentialSampler, RandomSampler
from torch.utils.data.distributed import DistributedSampler
from ..label_tensor import LabelTensor
from .dataset import PinaDatasetFactory, PinaTensorDataset


class DummyDataloader:

    def __init__(self, dataset):
        """
        Prepare a dataloader object that returns the entire dataset in a single
        batch. Depending on the number of GPUs, the dataset is managed
        as follows:

        - **Distributed Environment** (multiple GPUs): Divides dataset across
            processes using the rank and world size. Fetches only portion of
            data corresponding to the current process.
        - **Non-Distributed Environment** (single GPU): Fetches the entire
            dataset.

        :param PinaDataset dataset: The dataset object to be processed.

        .. note::
           This dataloader is used when the batch size is ``None``.
        """

        if (
            torch.distributed.is_available()
            and torch.distributed.is_initialized()
        ):
            rank = torch.distributed.get_rank()
            world_size = torch.distributed.get_world_size()
            if len(dataset) < world_size:
                raise RuntimeError(
                    "Dimension of the dataset smaller than world size."
                    " Increase the size of the partition or use a single GPU"
                )
            idx, i = [], rank
            while i < len(dataset):
                idx.append(i)
                i += world_size
            self.dataset = dataset.fetch_from_idx_list(idx)
        else:
            self.dataset = dataset.get_all_data()

    def __iter__(self):
        return self

    def __len__(self):
        return 1

    def __next__(self):
        return self.dataset


class Collator:
    """
    This callable class is used to collate the data points fetched from the
    dataset. The collation is performed based on the type of dataset used and
    on the batching strategy.
    """

    def __init__(
        self, max_conditions_lengths, automatic_batching, dataset=None
    ):
        """
        Initialize the object, setting the collate function based on whether
        automatic batching is enabled or not.

        :param dict max_conditions_lengths: ``dict`` containing the maximum
            number  of data points to consider in a single batch for
            each condition.
        :param bool automatic_batching: Whether automatic PyTorch batching is
            enabled or not. For more information, see the
            :class:`~pina.data.data_module.PinaDataModule` class.
        :param PinaDataset dataset: The dataset where the data is stored.
        """

        self.max_conditions_lengths = max_conditions_lengths
        # Set the collate function based on the batching strategy
        # collate_pina_dataloader is used when automatic batching is disabled
        # collate_torch_dataloader is used when automatic batching is enabled
        self.callable_function = (
            self._collate_torch_dataloader
            if automatic_batching
            else (self._collate_pina_dataloader)
        )
        self.dataset = dataset

        # Set the function which performs the actual collation
        if isinstance(self.dataset, PinaTensorDataset):
            # If the dataset is a PinaTensorDataset, use this collate function
            self._collate = self._collate_tensor_dataset
        else:
            # If the dataset is a PinaDataset, use this collate function
            self._collate = self._collate_graph_dataset

    def _collate_pina_dataloader(self, batch):
        """
        Function used to create a batch when automatic batching is disabled.

        :param list[int] batch: List of integers representing the indices of
            the data points to be fetched.
        :return: Dictionary containing the data points fetched from the dataset.
        :rtype: dict
        """
        # Call the fetch_from_idx_list method of the dataset
        return self.dataset.fetch_from_idx_list(batch)

    def _collate_torch_dataloader(self, batch):
        """
        Function used to collate the batch

        :param list[dict] batch: List of retrieved data.
        :return: Dictionary containing the data points fetched from the dataset,
            collated.
        :rtype: dict
        """

        batch_dict = {}
        if isinstance(batch, dict):
            return batch
        conditions_names = batch[0].keys()
        # Condition names
        for condition_name in conditions_names:
            single_cond_dict = {}
            condition_args = batch[0][condition_name].keys()
            for arg in condition_args:
                data_list = [
                    batch[idx][condition_name][arg]
                    for idx in range(
                        min(
                            len(batch),
                            self.max_conditions_lengths[condition_name],
                        )
                    )
                ]
                single_cond_dict[arg] = self._collate(data_list)

            batch_dict[condition_name] = single_cond_dict
        return batch_dict

    @staticmethod
    def _collate_tensor_dataset(data_list):
        """
        Function used to collate the data when the dataset is a
        :class:`~pina.data.dataset.PinaTensorDataset`.

        :param data_list: Elements to be collated.
        :type data_list: list[torch.Tensor] | list[LabelTensor]
        :return: Batch of data.
        :rtype: dict

        :raises RuntimeError: If the data is not a :class:`torch.Tensor` or a
            :class:`~pina.label_tensor.LabelTensor`.
        """

        if isinstance(data_list[0], LabelTensor):
            return LabelTensor.stack(data_list)
        if isinstance(data_list[0], torch.Tensor):
            return torch.stack(data_list)
        raise RuntimeError("Data must be Tensors or LabelTensor ")

    def _collate_graph_dataset(self, data_list):
        """
        Function used to collate data when the dataset is a
        :class:`~pina.data.dataset.PinaGraphDataset`.

        :param data_list: Elememts to be collated.
        :type data_list: list[Data] | list[Graph]
        :return: Batch of data.
        :rtype: dict

        :raises RuntimeError: If the data is not a
            :class:`~torch_geometric.data.Data` or a :class:`~pina.graph.Graph`.
        """
        if isinstance(data_list[0], LabelTensor):
            return LabelTensor.cat(data_list)
        if isinstance(data_list[0], torch.Tensor):
            return torch.cat(data_list)
        if isinstance(data_list[0], Data):
            return self.dataset.create_batch(data_list)
        raise RuntimeError(
            "Data must be Tensors or LabelTensor or pyG "
            "torch_geometric.data.Data"
        )

    def __call__(self, batch):
        """
        Perform the collation of data fetched from the dataset. The behavoior
        of the function is set based on the batching strategy during class
        initialization.

        :param batch: List of retrieved data or sampled indices.
        :type batch: list[int] | list[dict]
        :return: Dictionary containing colleted data fetched from the dataset.
        :rtype: dict
        """

        return self.callable_function(batch)


class PinaSampler:
    """
    This class is used to create the sampler instance based on the shuffle
    parameter and the environment in which the code is running.
    """

    def __new__(cls, dataset):
        """
        Instantiate and initialize the sampler.

        :param PinaDataset dataset: The dataset from which to sample.
        :return: The sampler instance.
        :rtype: :class:`torch.utils.data.Sampler`
        """

        if (
            torch.distributed.is_available()
            and torch.distributed.is_initialized()
        ):
            sampler = DistributedSampler(dataset)
        else:
            sampler = SequentialSampler(dataset)
        return sampler


class PinaDataModule(LightningDataModule):
    """
    This class extends :class:`~lightning.pytorch.core.LightningDataModule`,
    allowing proper creation and management of different types of datasets
    defined in PINA.
    """

    def __init__(
        self,
        problem,
        train_size=0.7,
        test_size=0.2,
        val_size=0.1,
        batch_size=None,
        shuffle=True,
        repeat=False,
        automatic_batching=None,
        num_workers=0,
        pin_memory=False,
    ):
        """
        Initialize the object and creating datasets based on the input problem.

        :param AbstractProblem problem: The problem containing the data on which
            to create the datasets and dataloaders.
        :param float train_size: Fraction of elements in the training split. It
            must be in the range [0, 1].
        :param float test_size: Fraction of elements in the test split. It must
            be in the range [0, 1].
        :param float val_size: Fraction of elements in the validation split. It
            must be in the range [0, 1].
        :param int batch_size: The batch size used for training. If ``None``,
            the entire dataset is returned in a single batch.
            Default is ``None``.
        :param bool shuffle: Whether to shuffle the dataset before splitting.
            Default ``True``.
        :param bool repeat: If ``True``, in case of batch size larger than the
            number of elements in a specific condition, the elements are
            repeated until the batch size is reached. If ``False``, the number
            of elements in the batch is the minimum between the batch size and
            the number of elements in the condition. Default is ``False``.
        :param automatic_batching: If ``True``, automatic PyTorch batching
            is performed, which consists of extracting one element at a time
            from the dataset and collating them into a batch. This is useful
            when the dataset is too large to fit into memory. On the other hand,
            if ``False``, the items are retrieved from the dataset all at once
            avoind the overhead of collating them into a batch and reducing the
            ``__getitem__`` calls to the dataset. This is useful when the
            dataset fits into memory. Avoid using automatic batching when
            ``batch_size`` is large. Default is ``False``.
        :param int num_workers: Number of worker threads for data loading.
            Default ``0`` (serial loading).
        :param bool pin_memory: Whether to use pinned memory for faster data
            transfer to GPU. Default ``False``.

        :raises ValueError: If at least one of the splits is negative.
        :raises ValueError: If the sum of the splits is different from 1.

        .. seealso::
            For more information on multi-process data loading, see:
            https://pytorch.org/docs/stable/data.html#multi-process-data-loading

            For details on memory pinning, see:
            https://pytorch.org/docs/stable/data.html#memory-pinning
        """
        super().__init__()

        # Store fixed attributes
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.repeat = repeat
        self.automatic_batching = automatic_batching

        # If batch size is None, num_workers has no effect
        if batch_size is None and num_workers != 0:
            warnings.warn(
                "Setting num_workers when batch_size is None has no effect on "
                "the DataLoading process."
            )
            self.num_workers = 0
        else:
            self.num_workers = num_workers

        # If batch size is None, pin_memory has no effect
        if batch_size is None and pin_memory:
            warnings.warn(
                "Setting pin_memory to True has no effect when "
                "batch_size is None."
            )
            self.pin_memory = False
        else:
            self.pin_memory = pin_memory

        # Collect data
        problem.collect_data()

        # Check if the splits are correct
        self._check_slit_sizes(train_size, test_size, val_size)

        # Split input data into subsets
        splits_dict = {}
        if train_size > 0:
            splits_dict["train"] = train_size
            self.train_dataset = None
        else:
            # Use the super method to create the train dataloader which
            # raises NotImplementedError
            self.train_dataloader = super().train_dataloader
        if test_size > 0:
            splits_dict["test"] = test_size
            self.test_dataset = None
        else:
            # Use the super method to create the train dataloader which
            # raises NotImplementedError
            self.test_dataloader = super().test_dataloader
        if val_size > 0:
            splits_dict["val"] = val_size
            self.val_dataset = None
        else:
            # Use the super method to create the train dataloader which
            # raises NotImplementedError
            self.val_dataloader = super().val_dataloader

        self.data_splits = self._create_splits(
            problem.collected_data, splits_dict
        )
        self.transfer_batch_to_device = self._transfer_batch_to_device

    def setup(self, stage=None):
        """
        Create the dataset objects for the given stage.
        If the stage is "fit", the training and validation datasets are created.
        If the stage is "test", the testing dataset is created.

        :param str stage: The stage for which to perform the dataset setup.

        :raises ValueError: If the stage is neither "fit" nor "test".
        """
        if stage == "fit" or stage is None:
            self.train_dataset = PinaDatasetFactory(
                self.data_splits["train"],
                max_conditions_lengths=self.find_max_conditions_lengths(
                    "train"
                ),
                automatic_batching=self.automatic_batching,
            )
            if "val" in self.data_splits.keys():
                self.val_dataset = PinaDatasetFactory(
                    self.data_splits["val"],
                    max_conditions_lengths=self.find_max_conditions_lengths(
                        "val"
                    ),
                    automatic_batching=self.automatic_batching,
                )
        elif stage == "test":
            self.test_dataset = PinaDatasetFactory(
                self.data_splits["test"],
                max_conditions_lengths=self.find_max_conditions_lengths("test"),
                automatic_batching=self.automatic_batching,
            )
        else:
            raise ValueError("stage must be either 'fit' or 'test'.")

    @staticmethod
    def _split_condition(single_condition_dict, splits_dict):
        """
        Split the condition into different stages.

        :param dict single_condition_dict: The condition to be split.
        :param dict splits_dict: The dictionary containing the number of
            elements in each stage.
        :return: A dictionary containing the split condition.
        :rtype: dict
        """

        len_condition = len(single_condition_dict["input"])

        lengths = [
            int(len_condition * length) for length in splits_dict.values()
        ]

        remainder = len_condition - sum(lengths)
        for i in range(remainder):
            lengths[i % len(lengths)] += 1

        splits_dict = {
            k: max(1, v) for k, v in zip(splits_dict.keys(), lengths)
        }
        to_return_dict = {}
        offset = 0

        for stage, stage_len in splits_dict.items():
            to_return_dict[stage] = {
                k: v[offset : offset + stage_len]
                for k, v in single_condition_dict.items()
                if k != "equation"
                # Equations are NEVER dataloaded
            }
            if offset + stage_len >= len_condition:
                offset = len_condition - 1
                continue
            offset += stage_len
        return to_return_dict

    def _create_splits(self, collector, splits_dict):
        """
        Create the dataset objects putting data in the correct splits.

        :param Collector collector: The collector object containing the data.
        :param dict splits_dict: The dictionary containing the number of
            elements in each stage.
        :return: The dictionary containing the dataset objects.
        :rtype: dict
        """

        # ----------- Auxiliary function ------------
        def _apply_shuffle(condition_dict, len_data):
            idx = torch.randperm(len_data)
            for k, v in condition_dict.items():
                if k == "equation":
                    continue
                if isinstance(v, list):
                    condition_dict[k] = [v[i] for i in idx]
                elif isinstance(v, LabelTensor):
                    condition_dict[k] = LabelTensor(v.tensor[idx], v.labels)
                elif isinstance(v, torch.Tensor):
                    condition_dict[k] = v[idx]
                else:
                    raise ValueError(f"Data type {type(v)} not supported")

        # ----------- End auxiliary function ------------

        split_names = list(splits_dict.keys())
        dataset_dict = {name: {} for name in split_names}
        for (
            condition_name,
            condition_dict,
        ) in collector.items():
            len_data = len(condition_dict["input"])
            if self.shuffle:
                _apply_shuffle(condition_dict, len_data)
            for key, data in self._split_condition(
                condition_dict, splits_dict
            ).items():
                dataset_dict[key].update({condition_name: data})
        return dataset_dict

    def _create_dataloader(self, split, dataset):
        """ "
        Create the dataloader for the given split.

        :param str split: The split on which to create the dataloader.
        :param str dataset: The dataset to be used for the dataloader.
        :return: The dataloader for the given split.
        :rtype: torch.utils.data.DataLoader
        """
        # Suppress the warning about num_workers.
        # In many cases, especially for PINNs,
        # serial data loading can outperform parallel data loading.
        warnings.filterwarnings(
            "ignore",
            message=(
                "The '(train|val|test)_dataloader' does not have many workers "
                "which may be a bottleneck."
            ),
            module="lightning.pytorch.trainer.connectors.data_connector",
        )
        # Use custom batching (good if batch size is large)
        if self.batch_size is not None:
            sampler = PinaSampler(dataset)
            if self.automatic_batching:
                collate = Collator(
                    self.find_max_conditions_lengths(split),
                    self.automatic_batching,
                    dataset=dataset,
                )
            else:
                collate = Collator(
                    None, self.automatic_batching, dataset=dataset
                )
            return DataLoader(
                dataset,
                self.batch_size,
                collate_fn=collate,
                sampler=sampler,
                num_workers=self.num_workers,
            )
        dataloader = DummyDataloader(dataset)
        dataloader.dataset = self._transfer_batch_to_device(
            dataloader.dataset, self.trainer.strategy.root_device, 0
        )
        self.transfer_batch_to_device = self._transfer_batch_to_device_dummy
        return dataloader

    def find_max_conditions_lengths(self, split):
        """
        Define the maximum length for each conditions.

        :param dict split: The split of the dataset.
        :return: The maximum length per condition.
        :rtype: dict
        """

        max_conditions_lengths = {}
        for k, v in self.data_splits[split].items():
            if self.batch_size is None:
                max_conditions_lengths[k] = len(v["input"])
            elif self.repeat:
                max_conditions_lengths[k] = self.batch_size
            else:
                max_conditions_lengths[k] = min(
                    len(v["input"]), self.batch_size
                )
        return max_conditions_lengths

    def val_dataloader(self):
        """
        Create the validation dataloader.

        :return: The validation dataloader
        :rtype: torch.utils.data.DataLoader
        """
        return self._create_dataloader("val", self.val_dataset)

    def train_dataloader(self):
        """
        Create the training dataloader

        :return: The training dataloader
        :rtype: torch.utils.data.DataLoader
        """
        return self._create_dataloader("train", self.train_dataset)

    def test_dataloader(self):
        """
        Create the testing dataloader

        :return: The testing dataloader
        :rtype: torch.utils.data.DataLoader
        """
        return self._create_dataloader("test", self.test_dataset)

    @staticmethod
    def _transfer_batch_to_device_dummy(batch, device, dataloader_idx):
        """
        Transfer the batch to the device. This method is used when the batch
        size is None: batch has already been transferred to the device.

        :param list[tuple] batch: List of tuple where the first element of the
            tuple is the condition name and the second element is the data.
        :param torch.device device: Device to which the batch is transferred.
        :param int dataloader_idx: Index of the dataloader.
        :return: The batch transferred to the device.
        :rtype: list[tuple]
        """

        return batch

    def _transfer_batch_to_device(self, batch, device, dataloader_idx):
        """
        Transfer the batch to the device. This method is called in the
        training loop and is used to transfer the batch to the device.

        :param dict batch: The batch to be transferred to the device.
        :param torch.device device: The device to which the batch is
            transferred.
        :param int dataloader_idx: The index of the dataloader.
        :return: The batch transferred to the device.
        :rtype: list[tuple]
        """

        batch = [
            (
                k,
                super(LightningDataModule, self).transfer_batch_to_device(
                    v, device, dataloader_idx
                ),
            )
            for k, v in batch.items()
        ]

        return batch

    @staticmethod
    def _check_slit_sizes(train_size, test_size, val_size):
        """
        Check if the splits are correct. The splits sizes must be positive and
        the sum of the splits must be 1.

        :param float train_size: The size of the training split.
        :param float test_size: The size of the testing split.
        :param float val_size: The size of the validation split.

        :raises ValueError: If at least one of the splits is negative.
        :raises ValueError: If the sum of the splits is different
            from 1.
        """

        if train_size < 0 or test_size < 0 or val_size < 0:
            raise ValueError("The splits must be positive")
        if abs(train_size + test_size + val_size - 1) > 1e-6:
            raise ValueError("The sum of the splits must be 1")

    @property
    def input(self):
        """
        Return all the input points coming from all the datasets.

        :return: The input points for training.
        :rtype: dict
        """

        to_return = {}
        if hasattr(self, "train_dataset") and self.train_dataset is not None:
            to_return["train"] = self.train_dataset.input
        if hasattr(self, "val_dataset") and self.val_dataset is not None:
            to_return["val"] = self.val_dataset.input
        if hasattr(self, "test_dataset") and self.test_dataset is not None:
            to_return["test"] = self.test_dataset.input
        return to_return



"""Module for the Causal PINN solver."""

import torch

from ...problem import TimeDependentProblem
from .pinn import PINN
from ...utils import check_consistency


class CausalPINN(PINN):
    r"""
    Causal Physics-Informed Neural Network (CausalPINN) solver class.
    This class implements the Causal Physics-Informed Neural Network solver,
    using a user specified ``model`` to solve a specific ``problem``.
    It can be used to solve both forward and inverse problems.

    The Causal Physics-Informed Neural Network solver aims to find the solution
    :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m` of a differential problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}

    minimizing the loss function:

    .. math::
        \mathcal{L}_{\rm{problem}} = \frac{1}{N_t}\sum_{i=1}^{N_t}
        \omega_{i}\mathcal{L}_r(t_i),

    where:

    .. math::
        \mathcal{L}_r(t) = \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathcal{A}[\mathbf{u}](\mathbf{x}_i, t)) +
        \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathcal{B}[\mathbf{u}](\mathbf{x}_i, t))

    and,

    .. math::
        \omega_i = \exp\left(\epsilon \sum_{k=1}^{i-1}\mathcal{L}_r(t_k)\right).

    :math:`\epsilon` is an hyperparameter, set by default to :math:`100`, while
    :math:`\mathcal{L}` is a specific loss function, typically the MSE:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.

    .. seealso::

        **Original reference**: Wang, Sifan, Shyam Sankaran, and Paris
        Perdikaris. 
        *Respecting causality for training physics-informed
        neural networks.*
        Computer Methods in Applied Mechanics and Engineering 421 (2024):116813.
        DOI: `10.1016 <https://doi.org/10.1016/j.cma.2024.116813>`_.

    .. note::
        This class is only compatible with problems that inherit from  the
        :class:`~pina.problem.time_dependent_problem.TimeDependentProblem`
        class.
    """

    def __init__(
        self,
        problem,
        model,
        optimizer=None,
        scheduler=None,
        weighting=None,
        loss=None,
        eps=100,
    ):
        """
        Initialization of the :class:`CausalPINN` class.

        :param AbstractProblem problem: The problem to be solved. It must
            inherit from at least
            :class:`~pina.problem.time_dependent_problem.TimeDependentProblem`.
        :param torch.nn.Module model: The neural network model to be used.
        :param Optimizer optimizer: The optimizer to be used.
            If ``None``, the :class:`torch.optim.Adam` optimizer is used.
            Default is ``None``.
        :param torch.optim.LRScheduler scheduler: Learning rate scheduler.
            If ``None``, the :class:`torch.optim.lr_scheduler.ConstantLR`
            scheduler is used. Default is ``None``.
        :param WeightingInterface weighting: The weighting schema to be used.
            If ``None``, no weighting schema is used. Default is ``None``.
        :param torch.nn.Module loss: The loss function to be minimized.
            If ``None``, the :class:`torch.nn.MSELoss` loss is used.
            Default is `None`.
        :param float eps: The exponential decay parameter. Default is ``100``.
        :raises ValueError: If the problem is not a TimeDependentProblem.
        """
        super().__init__(
            model=model,
            problem=problem,
            optimizer=optimizer,
            scheduler=scheduler,
            weighting=weighting,
            loss=loss,
        )

        # checking consistency
        check_consistency(eps, (int, float))
        self._eps = eps
        if not isinstance(self.problem, TimeDependentProblem):
            raise ValueError(
                "Casual PINN works only for problems"
                "inheriting from TimeDependentProblem."
            )

    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the physics-informed solver based on the
        provided samples and equation.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation.
        :return: The computed physics loss.
        :rtype: LabelTensor
        """
        # split sequentially ordered time tensors into chunks
        chunks, labels = self._split_tensor_into_chunks(samples)
        # compute residuals - this correspond to ordered loss functions
        # values for each time step. Apply `flatten` to ensure obtaining
        # a tensor of shape #chunks after concatenating the residuals
        time_loss = []
        for chunk in chunks:
            chunk.labels = labels
            # classical PINN loss
            residual = self.compute_residual(samples=chunk, equation=equation)
            loss_val = self._loss_fn(
                torch.zeros_like(residual, requires_grad=True), residual
            )
            time_loss.append(loss_val)

        # concatenate residuals
        time_loss = torch.stack(time_loss)
        # compute weights without storing the gradient
        with torch.no_grad():
            weights = self._compute_weights(time_loss)
        return (weights * time_loss).mean()

    @property
    def eps(self):
        """
        The exponential decay parameter.

        :return: The exponential decay parameter.
        :rtype: float
        """
        return self._eps

    @eps.setter
    def eps(self, value):
        """
        Set the exponential decay parameter.

        :param float value: The exponential decay parameter.
        """
        check_consistency(value, float)
        self._eps = value

    def _sort_label_tensor(self, tensor):
        """
        Sort the tensor with respect to the temporal variables.

        :param LabelTensor tensor: The tensor to be sorted.
        :return: The tensor sorted with respect to the temporal variables.
        :rtype: LabelTensor
        """
        # labels input tensors
        labels = tensor.labels
        # extract time tensor
        time_tensor = tensor.extract(self.problem.temporal_domain.variables)
        # sort the time tensors (this is very bad for GPU)
        _, idx = torch.sort(time_tensor.tensor.flatten())
        tensor = tensor[idx]
        tensor.labels = labels
        return tensor

    def _split_tensor_into_chunks(self, tensor):
        """
        Split the tensor into chunks based on time.

        :param LabelTensor tensor: The tensor to be split.
        :return: A tuple containing the list of tensor chunks and the
            corresponding labels.
        :rtype: tuple[list[LabelTensor], list[str]]
        """
        # extract labels
        labels = tensor.labels
        # sort input tensor based on time
        tensor = self._sort_label_tensor(tensor)
        # extract time tensor
        time_tensor = tensor.extract(self.problem.temporal_domain.variables)
        # count unique tensors in time
        _, idx_split = time_tensor.unique(return_counts=True)
        # split the tensor based on time
        chunks = torch.split(tensor, tuple(idx_split))
        return chunks, labels

    def _compute_weights(self, loss):
        """
        Compute the weights for the physics loss based on the cumulative loss.

        :param LabelTensor loss: The physics loss values.
        :return: The computed weights for the physics loss.
        :rtype: LabelTensor
        """
        # compute comulative loss and multiply by epsilon
        cumulative_loss = self._eps * torch.cumsum(loss, dim=0)
        # return the exponential of the negative weighted cumulative sum
        return torch.exp(-cumulative_loss)



"""Module for Neural Tangent Kernel Class"""

import torch
from torch.nn import Module
from .weighting_interface import WeightingInterface
from ..utils import check_consistency


class NeuralTangentKernelWeighting(WeightingInterface):
    """
    A neural tangent kernel scheme for weighting different losses to
    boost the convergence.

    .. seealso::

        **Original reference**: Wang, Sifan, Xinling Yu, and
        Paris Perdikaris. *When and why PINNs fail to train:
        A neural tangent kernel perspective*. Journal of
        Computational Physics 449 (2022): 110768.
        DOI: `10.1016 <https://doi.org/10.1016/j.jcp.2021.110768>`_.

    """

    def __init__(self, model, alpha=0.5):
        """
        Initialization of the :class:`NeuralTangentKernelWeighting` class.

        :param torch.nn.Module model: The neural network model.
        :param float alpha: The alpha parameter.

        :raises ValueError: If ``alpha`` is not between 0 and 1 (inclusive).
        """

        super().__init__()
        check_consistency(alpha, float)
        check_consistency(model, Module)
        if alpha < 0 or alpha > 1:
            raise ValueError("alpha should be a value between 0 and 1")
        self.alpha = alpha
        self.model = model
        self.weights = {}
        self.default_value_weights = 1

    def aggregate(self, losses):
        """
        Weight the losses according to the Neural Tangent Kernel
        algorithm.

        :param dict(torch.Tensor) input: The dictionary of losses.
        :return: The losses aggregation. It should be a scalar Tensor.
        :rtype: torch.Tensor
        """
        losses_norm = {}
        for condition in losses:
            losses[condition].backward(retain_graph=True)
            grads = []
            for param in self.model.parameters():
                grads.append(param.grad.view(-1))
            grads = torch.cat(grads)
            losses_norm[condition] = torch.norm(grads)
        self.weights = {
            condition: self.alpha
            * self.weights.get(condition, self.default_value_weights)
            + (1 - self.alpha)
            * losses_norm[condition]
            / sum(losses_norm.values())
            for condition in losses
        }
        return sum(
            self.weights[condition] * loss for condition, loss in losses.items()
        )



"""Module for the Low Rank Neural Operator model class."""

import torch
from torch import nn

from ..utils import check_consistency

from .kernel_neural_operator import KernelNeuralOperator
from .block.low_rank_block import LowRankBlock


class LowRankNeuralOperator(KernelNeuralOperator):
    """
    Low Rank Neural Operator model class.

    The Low Rank Neural Operator is a general architecture for learning
    operators, which map functions to functions. It can be trained both with
    Supervised and Physics-Informed learning strategies. The Low Rank Neural
    Operator performs convolution by means of a low rank approximation.

    .. seealso::

        **Original reference**: Kovachki, N., Li, Z., Liu, B., Azizzadenesheli,
        K., Bhattacharya, K., Stuart, A., & Anandkumar, A. (2023).
        *Neural operator: Learning maps between function spaces with
        applications to PDEs*.
        Journal of Machine Learning Research, 24(89), 1-97.
    """

    def __init__(
        self,
        lifting_net,
        projecting_net,
        field_indices,
        coordinates_indices,
        n_kernel_layers,
        rank,
        inner_size=20,
        n_layers=2,
        func=torch.nn.Tanh,
        bias=True,
    ):
        """
        Initialization of the :class:`LowRankNeuralOperator` class.

        :param torch.nn.Module lifting_net: The lifting neural network mapping
            the input to its hidden dimension. It must take as input the input
            field and the coordinates at which the input field is evaluated.
        :param torch.nn.Module projecting_net: The projection neural network
            mapping the hidden representation to the output function. It must
            take as input the embedding dimension plus the dimension of the
            coordinates.
        :param list[str] field_indices: The labels of the fields in the input
            tensor.
        :param list[str] coordinates_indices: The labels of the coordinates in
            the input tensor.
        :param int n_kernel_layers: The number of hidden kernel layers.
        :param int rank: The rank of the low rank approximation.
        :param int inner_size: The number of neurons for each hidden layer in
            the basis function neural network. Default is ``20``.
        :param int n_layers: The number of hidden layers in the basis function
            neural network. Default is ``2``.
        :param func: The activation function. If a list is passed, it must have
            the same length as ``n_layers``. If a single function is passed, it
            is used for all layers, except for the last one.
            Default is :class:`torch.nn.Tanh`.
        :type func: torch.nn.Module | list[torch.nn.Module]
        :param bool bias: If ``True`` bias is considered for the basis function
            neural network. Default is ``True``.
        :raises ValueError: If the input dimension does not match with the
            labels of the fields and coordinates.
        :raises ValueError: If the input dimension of the projecting network
            does not match with the hidden dimension of the lifting network.
        """

        # check consistency
        check_consistency(field_indices, str)
        check_consistency(coordinates_indices, str)
        check_consistency(n_kernel_layers, int)

        # check hidden dimensions match
        input_lifting_net = next(lifting_net.parameters()).size()[-1]
        output_lifting_net = lifting_net(
            torch.rand(size=next(lifting_net.parameters()).size())
        ).shape[-1]
        projecting_net_input = next(projecting_net.parameters()).size()[-1]

        if len(field_indices) + len(coordinates_indices) != input_lifting_net:
            raise ValueError(
                "The lifting_net must take as input the "
                "coordinates vector and the field vector."
            )

        if (
            output_lifting_net + len(coordinates_indices)
            != projecting_net_input
        ):
            raise ValueError(
                "The projecting_net input must be equal to "
                "the embedding dimension (which is the output) "
                "of the lifting_net plus the dimension of the "
                "coordinates, i.e. len(coordinates_indices)."
            )

        # assign
        self.coordinates_indices = coordinates_indices
        self.field_indices = field_indices
        integral_net = nn.Sequential(
            *[
                LowRankBlock(
                    input_dimensions=len(coordinates_indices),
                    embedding_dimenion=output_lifting_net,
                    rank=rank,
                    inner_size=inner_size,
                    n_layers=n_layers,
                    func=func,
                    bias=bias,
                )
                for _ in range(n_kernel_layers)
            ]
        )
        super().__init__(lifting_net, integral_net, projecting_net)

    def forward(self, x):
        r"""
        Forward pass for the :class:`LowRankNeuralOperator` model.

        The ``lifting_net`` maps the input to the hidden dimension.
        Then, several layers of
        :class:`~pina.model.block.low_rank_block.LowRankBlock` are
        applied. Finally, the ``projecting_net`` maps the hidden representation
        to the output function.

        :param LabelTensor x: The input tensor for performing the computation.
            It expects a tensor :math:`B \times N \times D`, where :math:`B` is
            the batch_size, :math:`N` the number of points in the mesh,
            :math:`D` the dimension of the problem, i.e. the sum
            of ``len(coordinates_indices)`` and ``len(field_indices)``.
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        # extract points
        coords = x.extract(self.coordinates_indices)
        # lifting
        x = self._lifting_operator(x)
        # kernel
        for module in self._integral_kernels:
            x = module(x, coords)
        # projecting
        return self._projection_operator(torch.cat((x, coords), dim=-1))



"""Module to define equations and systems of equations."""

__all__ = [
    "SystemEquation",
    "Equation",
    "FixedValue",
    "FixedGradient",
    "FixedFlux",
    "Laplace",
]

from .equation import Equation
from .equation_factory import FixedFlux, FixedGradient, Laplace, FixedValue
from .system_equation import SystemEquation
